from __future__ import annotations

import shlex
from collections import defaultdict
from io import StringIO
from typing import Callable, NamedTuple, cast
from urllib.parse import urlparse

from packaging.requirements import InvalidRequirement, Requirement

from pyinfra import logger
from pyinfra.api import Host, OperationValueError, State
from pyinfra.facts.files import File
from pyinfra.facts.rpm import RpmPackage
from pyinfra.operations import files


def default_inst_vers_format_fn(name: str, operator: str, version: str):
    return "{name}{operator}{version}".format(name=name, operator=operator, version=version)


class PkgInfo(NamedTuple):
    name: str
    version: str
    operator: str
    url: str
    inst_vers_format_fn: Callable = default_inst_vers_format_fn
    """
    The key packaging information needed: version, operator and url are optional.
    """

    @property
    def lkup_name(self) -> str | list[str]:
        return self.name if self.version == "" else [self.name, self.version]

    @property
    def has_version(self) -> bool:
        return self.version != ""

    @property
    def inst_vers(self) -> str:
        """String that represents how a program can be installed.

        - If self.url exists, then url is always returned.
        - If self.version exists, then inst_vers_format_fn is used
        to create the string. The default template is '{name}{operator}{version}'.
        - Otherwise, self.name is returned.

        Note, the result string will be quoted, so input is shell safe.
        """

        if self.url:
            return shlex.quote(self.url)

        if self.version:
            return shlex.quote(self.inst_vers_format_fn(self.name, self.operator, self.version))
        return shlex.quote(self.name)

    @classmethod
    def from_possible_pair(cls, s: str, join: str | None) -> PkgInfo:
        if join is not None:
            pieces = s.rsplit(join, 1)
            return cls(pieces[0], pieces[1] if len(pieces) > 1 else "", join, "")

        return cls(s, "", "", "")

    @classmethod
    def from_pep508(cls, s: str) -> PkgInfo | None:
        """
        Separate out the useful parts (name, url, operator, version) of a PEP-508 dependency.
        Note: only one specifier is allowed.
        PEP-0426 states that Python packages should be compared using lowercase; thus
        the name is lower-cased
        For backwards compatibility, invalid requirements are assumed to be package names with a
        warning that this will change in the next major release
        """
        pep_508 = "PEP 508 non-compliant "
        treatment = "requirement treated as package name"
        will_change = "4.x will make this an error"  # pip and pipx already throw away None's
        try:
            reqt = Requirement(s)
        except InvalidRequirement as e:
            logger.warning(f"{pep_508} :{e}\n{will_change}")
            return cls(s, "", "", "")
        else:
            if (len(reqt.specifier) > 0) and (len(reqt.specifier) > 1):
                logger.warning(f"{pep_508}/unsupported specifier ({s}) {treatment}\n{will_change}")
                return cls(s, "", "", "")
            else:
                spec = next(iter(reqt.specifier), None)
                return cls(
                    reqt.name.lower(),
                    spec.version if spec is not None else "",
                    spec.operator if spec is not None else "",
                    reqt.url or "",
                )


def _has_package(
    package: str | list[str],
    packages: dict[str, set[str]],
    expand_package_fact: Callable[[str], list[str | list[str]]] | None = None,
    match_any=False,
) -> tuple[bool, dict]:
    def in_packages(pkg_name, pkg_versions):
        if not pkg_versions:
            return pkg_name in packages
        return pkg_name in packages and any(
            version in packages[pkg_name] for version in pkg_versions
        )

    packages_to_check: list[str | list[str]] = [package]
    if expand_package_fact:
        if isinstance(package, list):
            packages_to_check = expand_package_fact(package[0]) or packages_to_check
        else:
            packages_to_check = expand_package_fact(package) or packages_to_check

    package_name_to_versions = defaultdict(set)
    for pkg in packages_to_check:
        if isinstance(pkg, list):
            package_name_to_versions[pkg[0]].add(pkg[1])
        else:
            package_name_to_versions[pkg]  # just make sure it exists

    checks = (
        in_packages(pkg_name, pkg_versions)
        for pkg_name, pkg_versions in package_name_to_versions.items()
    )

    if match_any:
        return any(checks), package_name_to_versions
    return all(checks), package_name_to_versions


def ensure_packages(
    host: Host,
    packages_to_ensure: str | list[str] | list[PkgInfo] | None,
    current_packages: dict[str, set[str]],
    present: bool,
    install_command: str,
    uninstall_command: str,
    latest: bool = False,
    upgrade_command: str | None = None,
    version_join: str | None = None,
    expand_package_fact: Callable[[str], list[str | list[str]]] | None = None,
):
    """
    Handles this common scenario:

    + We have a list of packages(/versions/urls) to ensure
    + We have a map of existing package -> versions
    + We have the common command bits (install, uninstall, version "joiner")
    + Outputs commands to ensure our desired packages/versions
    + Optionally upgrades packages w/o specified version when present

    Args:
        packages_to_ensure (list): list of packages or package/versions or PkgInfo's
        current_packages (dict): dict of package names -> version
        present (bool): whether packages should exist or not
        install_command (str): command to prefix to list of packages to install
        uninstall_command (str): as above for uninstalling packages
        latest (bool): whether to upgrade installed packages when present
        upgrade_command (str): as above for upgrading
        version_join (str): the package manager specific "joiner", ie ``=`` for \
            ``<apt_pkg>=<version>``.  Not allowed if (pkg, ver, url) tuples are provided.
        expand_package_fact: fact returning packages providing a capability \
            (ie ``yum whatprovides``)
    """

    if packages_to_ensure is None:
        return
    if isinstance(packages_to_ensure, str):
        packages_to_ensure = [packages_to_ensure]
    if len(packages_to_ensure) == 0:
        return

    packages: list[PkgInfo] = []
    if isinstance(packages_to_ensure[0], PkgInfo):
        packages = cast("list[PkgInfo]", packages_to_ensure)
        if version_join is not None:
            raise OperationValueError("cannot specify version_join and provide list[PkgInfo]")
    else:
        packages = [
            PkgInfo.from_possible_pair(package, version_join)
            for package in cast("list[str]", packages_to_ensure)
        ]

    diff_packages = []
    diff_expanded_packages = {}

    upgrade_packages = []

    if present is True:
        for package in packages:
            has_package, expanded_packages = _has_package(
                package.lkup_name, current_packages, expand_package_fact
            )

            if not has_package:
                diff_packages.append(package.inst_vers)
                diff_expanded_packages[package.name] = expanded_packages
            else:
                # Present packages w/o version specified - for upgrade if latest
                if not package.has_version:  # don't try to upgrade if a specific version requested
                    upgrade_packages.append(package.inst_vers)

                if not latest:
                    if (pkg := package.name) in current_packages:
                        host.noop(f"package {pkg} is installed ({','.join(current_packages[pkg])})")
                    else:
                        host.noop(f"package {package.name} is installed")
    if present is False:
        for package in packages:
            has_package, expanded_packages = _has_package(
                package.lkup_name, current_packages, expand_package_fact, match_any=True
            )

            if has_package:
                diff_packages.append(package.inst_vers)
                diff_expanded_packages[package.name] = expanded_packages
            else:
                host.noop(f"package {package.name} is not installed")

    if diff_packages:
        command = install_command if present else uninstall_command
        yield f"{command} {' '.join([pkg for pkg in diff_packages])}"

    if latest and upgrade_command and upgrade_packages:
        yield f"{upgrade_command} {' '.join([pkg for pkg in upgrade_packages])}"


def ensure_rpm(state: State, host: Host, source: str, present: bool, package_manager_command: str):
    original_source = source

    # If source is a url
    if urlparse(source).scheme:
        # Generate a temp filename (with .rpm extension to please yum)
        temp_filename = "{0}.rpm".format(host.get_temp_filename(source))

        # Ensure it's downloaded
        yield from files.download._inner(src=source, dest=temp_filename)

        # Override the source with the downloaded file
        source = temp_filename

    # Check for file .rpm information
    info = host.get_fact(RpmPackage, package=source)
    exists = False

    # We have info!
    if info:
        current_package = host.get_fact(RpmPackage, package=info["name"])
        if current_package and current_package["version"] == info["version"]:
            exists = True

    # Package does not exist and we want?
    if present and not exists:
        # If we had info, always install
        if info:
            yield "rpm -i {0}".format(source)
        # This happens if we download the package mid-deploy, so we have no info
        # but also don't know if it's installed. So check at runtime, otherwise
        # the install will fail.
        else:
            yield "rpm -q `rpm -qp {0}` 2> /dev/null || rpm -i {0}".format(source)

    # Package exists but we don't want?
    elif exists and not present:
        yield "{0} remove -y {1}".format(package_manager_command, info["name"])
    else:
        host.noop(
            "rpm {0} is {1}".format(
                original_source,
                "installed" if present else "not installed",
            ),
        )


def ensure_yum_repo(
    host: Host,
    name_or_url: str,
    baseurl: str | None,
    present: bool,
    description: str | None,
    enabled: bool,
    gpgcheck: bool,
    gpgkey: str | None,
    repo_directory="/etc/yum.repos.d/",
    type_: str | None = None,
):
    url = None
    url_parts = urlparse(name_or_url)
    if url_parts.scheme:
        url = name_or_url
        name_or_url = url_parts.path.split("/")[-1]
        if name_or_url.endswith(".repo"):
            name_or_url = name_or_url[:-5]

    filename = "{0}{1}.repo".format(repo_directory, name_or_url)

    # If we don't want the repo, just remove any existing file
    if not present:
        yield from files.file._inner(path=filename, present=False)
        return

    # If we're a URL, download the repo if it doesn't exist
    if url:
        if not host.get_fact(File, path=filename):
            yield from files.download._inner(src=url, dest=filename)
        return

    assert isinstance(baseurl, str)

    # Description defaults to name
    description = description or name_or_url

    # Build the repo file from string
    repo_lines = [
        "[{0}]".format(name_or_url),
        "name={0}".format(description),
        "baseurl={0}".format(baseurl),
        "enabled={0}".format(1 if enabled else 0),
        "gpgcheck={0}".format(1 if gpgcheck else 0),
    ]

    if type_:
        repo_lines.append("type={0}".format(type_))

    if gpgkey:
        repo_lines.append("gpgkey={0}".format(gpgkey))

    repo_lines.append("")
    repo = "\n".join(repo_lines)
    repo_file = StringIO(repo)

    # Ensure this is the file on the server
    yield from files.put._inner(src=repo_file, dest=filename)



from __future__ import annotations

import json
from typing import Any, Dict, Iterable, List, TypeVar

from typing_extensions import override

from pyinfra.api import FactBase

T = TypeVar("T")


class PodmanFactBase(FactBase[T]):
    """
    Base for facts using `podman` to retrieve
    """

    abstract = True

    @override
    def requires_command(self, *args, **kwargs) -> str:
        return "podman"


class PodmanSystemInfo(PodmanFactBase[Dict[str, Any]]):
    """
    Output of 'podman system info'
    """

    @override
    def command(self) -> str:
        return "podman system info --format=json"

    @override
    def process(self, output: Iterable[str]) -> Dict[str, Any]:
        output = json.loads(("").join(output))
        assert isinstance(output, dict)
        return output


class PodmanPs(PodmanFactBase[List[Dict[str, Any]]]):
    """
    Output of 'podman ps'
    """

    @override
    def command(self) -> str:
        return "podman ps --format=json --all"

    @override
    def process(self, output: Iterable[str]) -> List[Dict[str, Any]]:
        output = json.loads(("").join(output))
        assert isinstance(output, list)
        return output  # type: ignore



from os import path

from pyinfra import local
from pyinfra.operations import server

server.shell(
    name="First task operation",
    commands="echo first_task_operation",
)

for item in [1, 2]:
    server.shell(
        name="Task order loop {0}".format(item),
        commands="echo loop_{0}".format(item),
    )
    server.shell(
        name="2nd Task order loop {0}".format(item),
        commands="echo loop_{0}".format(item),
    )

# Import a file *relative* to this one (./empty_task.py)
local.include(path.join(".", "nested", "empty_task.py"))

# Import a file from the CWD (tasks/another_task.py)
local.include(path.join("tasks", "another_task.py"))



from typing_extensions import override

from pyinfra.api import FactBase


class RunitStatus(FactBase):
    """
    Returns a dict of name -> status for runit services.

    + service: optionally check only for a single service
    + svdir: alternative ``SVDIR``

    .. code:: python

        {
            'agetty-tty1': True,     # service is running
            'dhcpcd': False,         # service is down
            'wpa_supplicant': None,  # service is managed, but not running or down,
                                     # possibly in a fail state
        }
    """

    default = dict

    @override
    def requires_command(self, *args, **kwargs) -> str:
        return "sv"

    @override
    def command(self, service=None, svdir="/var/service") -> str:
        if service is None:
            return (
                'export SVDIR="{0}" && '
                'cd "$SVDIR" && find * -maxdepth 0 -exec sv status {{}} + 2>/dev/null'
            ).format(svdir)
        else:
            return 'SVDIR="{0}" sv status "{1}"'.format(svdir, service)

    @override
    def process(self, output):
        services = {}
        for line in output:
            statusstr, service, _ = line.split(sep=": ", maxsplit=2)
            status = None

            if statusstr == "run":
                status = True
            elif statusstr == "down":
                status = False
            # another observable state is "fail"
            # report as ``None`` for now

            services[service] = status

        return services


class RunitManaged(FactBase):
    """
    Returns a set of all services managed by runit

    + service: optionally check only for a single service
    + svdir: alternative ``SVDIR``
    """

    default = set

    @override
    def command(self, service=None, svdir="/var/service"):
        if service is None:
            return 'cd "{0}" && find -mindepth 1 -maxdepth 1 -type l -printf "%f\n"'.format(svdir)
        else:
            return 'cd "{0}" && test -h "{1}" && echo "{1}" || true'.format(svdir, service)

    @override
    def process(self, output):
        return set(output)



"""
Facts about Docker containers, volumes and networks. These facts give you information from the view
of the current inventory host. See the :doc:`../connectors/docker` to use Docker containers as
inventory directly.
"""

from __future__ import annotations

import json

from typing_extensions import override

from pyinfra.api import FactBase


class DockerFactBase(FactBase):
    abstract = True

    docker_type: str

    @override
    def requires_command(self, *args, **kwargs) -> str:
        return "docker"

    @override
    def process(self, output):
        output = "".join(output)
        return json.loads(output)


class DockerSystemInfo(DockerFactBase):
    """
    Returns ``docker system info`` output in JSON format.
    """

    @override
    def command(self) -> str:
        return 'docker system info --format="{{json .}}"'


# All Docker objects
#


class DockerContainers(DockerFactBase):
    """
    Returns ``docker inspect`` output for all Docker containers.
    """

    @override
    def command(self) -> str:
        return """
        ids=$(docker ps -qa) && [ -n "$ids" ] && docker container inspect $ids || echo "[]"
        """.strip()


class DockerImages(DockerFactBase):
    """
    Returns ``docker inspect`` output for all Docker images.
    """

    @override
    def command(self) -> str:
        return """
        ids=$(docker images -q) && [ -n "$ids" ] && docker image inspect $ids || echo "[]"
        """.strip()


class DockerNetworks(DockerFactBase):
    """
    Returns ``docker inspect`` output for all Docker networks.
    """

    @override
    def command(self) -> str:
        return "docker network inspect `docker network ls -q`"


class DockerVolumes(DockerFactBase):
    """
    Returns ``docker inspect`` output for all Docker volumes.
    """

    @override
    def command(self) -> str:
        return "docker volume inspect `docker volume ls -q`"


class DockerPlugins(DockerFactBase):
    """
    Returns ``docker plugin inspect`` output for all Docker plugins.
    """

    @override
    def command(self) -> str:
        return """
        ids=$(docker plugin ls -q) && [ -n "$ids" ] && docker plugin inspect $ids || echo "[]"
        """.strip()


# Single Docker objects
#


class DockerSingleMixin(DockerFactBase):
    @override
    def command(self, object_id):
        return "docker {0} inspect {1} 2>&- || true".format(
            self.docker_type,
            object_id,
        )


class DockerContainer(DockerSingleMixin):
    """
    Returns ``docker inspect`` output for a single Docker container.
    """

    docker_type = "container"


class DockerImage(DockerSingleMixin):
    """
    Returns ``docker inspect`` output for a single Docker image.
    """

    docker_type = "image"


class DockerNetwork(DockerSingleMixin):
    """
    Returns ``docker inspect`` output for a single Docker network.
    """

    docker_type = "network"


class DockerVolume(DockerSingleMixin):
    """
    Returns ``docker inspect`` output for a single Docker container.
    """

    docker_type = "volume"


class DockerPlugin(DockerSingleMixin):
    """
    Returns ``docker plugin inspect`` output for a single Docker plugin.
    """

    docker_type = "plugin"



from __future__ import annotations

from typing_extensions import override

from .sysvinit import InitdStatus


class RcdStatus(InitdStatus):
    """
    Same as ``initd_status`` but for BSD (/etc/rc.d) systems. Unlike Linux/init.d,
    BSD init scripts are well behaved and as such their output can be trusted.
    """

    @override
    def command(self) -> str:
        return """
            for SERVICE in `find /etc/rc.d /usr/local/etc/rc.d -type f`; do
                $SERVICE status 2> /dev/null || $SERVICE check 2> /dev/null
                echo "`basename $SERVICE`=$?"
            done
        """

    default = dict



"""
The pyinfra facts API. Facts enable pyinfra to collect remote server state which
is used to "diff" with the desired state, producing the final commands required
for a deploy.

Note that the facts API does *not* use the global currently in context host so
it's possible to call facts on hosts out of context (ie give me the IP of this
other host B while I operate on this host A).
"""

from __future__ import annotations

import inspect
import re
from inspect import getcallargs
from socket import error as socket_error, timeout as timeout_error
from typing import TYPE_CHECKING, Any, Callable, Generic, Optional, Type, TypeVar, cast

import click
import gevent
from paramiko import SSHException
from typing_extensions import override

from pyinfra import logger
from pyinfra.api import StringCommand
from pyinfra.api.arguments import all_global_arguments, pop_global_arguments
from pyinfra.api.exceptions import FactProcessError
from pyinfra.api.util import (
    get_kwargs_str,
    log_error_or_warning,
    log_host_command_error,
    print_host_combined_output,
)
from pyinfra.connectors.util import CommandOutput
from pyinfra.context import ctx_host, ctx_state
from pyinfra.progress import progress_spinner

from .arguments import CONNECTOR_ARGUMENT_KEYS

if TYPE_CHECKING:
    from pyinfra.api import Host, State

SUDO_REGEX = r"^sudo: unknown user"
SU_REGEXES = (
    r"^su: user .+ does not exist",
    r"^su: unknown login",
)


T = TypeVar("T")


class FactBase(Generic[T]):
    name: str

    abstract: bool = True

    shell_executable: str | None = None

    command: Callable[..., str | StringCommand]

    def requires_command(self, *args, **kwargs) -> str | None:
        return None

    @override
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        module_name = cls.__module__.replace("pyinfra.facts.", "")
        cls.name = f"{module_name}.{cls.__name__}"

        # Check that fact's `command` method does not inadvertently take a global
        # argument, most commonly `name`.
        if hasattr(cls, "command") and callable(cls.command):
            command_args = set(inspect.signature(cls.command).parameters.keys())
            global_args = set([name for name, _ in all_global_arguments()])
            command_global_args = command_args & global_args

            if len(command_global_args) > 0:
                names = ", ".join(command_global_args)
                raise TypeError(f"{cls.name}'s arguments {names} are reserved for global arguments")

    @staticmethod
    def default() -> T:
        """
        Set the default attribute to be a type (eg list/dict).
        """

        return cast(T, None)

    def process(self, output: list[str]) -> T:
        # NOTE: TypeVar does not support a default, so we have to cast this str -> T
        return cast(T, "\n".join(output))

    def process_pipeline(self, args, output):
        return {arg: self.process([output[i]]) for i, arg in enumerate(args)}


class ShortFactBase(Generic[T]):
    name: str
    fact: Type[FactBase]

    @override
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        module_name = cls.__module__.replace("pyinfra.facts.", "")
        cls.name = f"{module_name}.{cls.__name__}"

    def process_data(self, data):
        return data


def get_short_facts(state: "State", host: "Host", short_fact, **kwargs):
    fact_data = get_fact(state, host, short_fact.fact, **kwargs)
    return short_fact().process_data(fact_data)


def _make_command(command_attribute, host_args):
    if callable(command_attribute):
        host_args.pop("self", None)
        return command_attribute(**host_args)
    return command_attribute


def _handle_fact_kwargs(state: "State", host: "Host", cls, args, kwargs):
    args = args or []
    kwargs = kwargs or {}

    # Start with a (shallow) copy of current operation kwargs if any
    ctx_kwargs: dict[str, Any] = (
        cast(dict[str, Any], host.current_op_global_arguments) or {}
    ).copy()
    # Update with the input kwargs (overrides)
    ctx_kwargs.update(kwargs)

    # Pop executor kwargs, pass remaining
    global_kwargs, _ = pop_global_arguments(state, host, cast(dict[str, Any], ctx_kwargs))

    fact_kwargs = {key: value for key, value in kwargs.items() if key not in global_kwargs}

    if args or fact_kwargs:
        # Merges args & kwargs into a single kwargs dictionary
        fact_kwargs = getcallargs(cls().command, *args, **fact_kwargs)

    return fact_kwargs, global_kwargs


def get_facts(state, *args, **kwargs):
    def get_host_fact(host, *args, **kwargs):
        with ctx_host.use(host):
            return get_fact(state, host, *args, **kwargs)

    with ctx_state.use(state):
        greenlet_to_host = {
            state.pool.spawn(get_host_fact, host, *args, **kwargs): host
            for host in state.inventory.get_active_hosts()
        }

    results = {}

    with progress_spinner(greenlet_to_host.values()) as progress:
        for greenlet in gevent.iwait(greenlet_to_host.keys()):
            host = greenlet_to_host[greenlet]
            results[host] = greenlet.get()
            progress(host)

    return results


def get_fact(
    state: "State",
    host: "Host",
    cls: type[FactBase],
    args: Optional[Any] = None,
    kwargs: Optional[Any] = None,
    ensure_hosts: Optional[Any] = None,
    apply_failed_hosts: bool = True,
) -> Any:
    if issubclass(cls, ShortFactBase):
        return get_short_facts(
            state,
            host,
            cls,
            args=args,
            kwargs=kwargs,
            ensure_hosts=ensure_hosts,
            apply_failed_hosts=apply_failed_hosts,
        )

    return _get_fact(
        state,
        host,
        cls,
        args,
        kwargs,
        ensure_hosts,
        apply_failed_hosts,
    )


def _get_fact(
    state: "State",
    host: "Host",
    cls: type[FactBase],
    args: Optional[list] = None,
    kwargs: Optional[dict] = None,
    ensure_hosts: Optional[Any] = None,
    apply_failed_hosts: bool = True,
) -> Any:
    fact = cls()
    name = fact.name

    fact_kwargs, global_kwargs = _handle_fact_kwargs(state, host, cls, args, kwargs)

    kwargs_str = get_kwargs_str(fact_kwargs)
    logger.debug(
        "Getting fact: %s (%s) (ensure_hosts: %r)",
        name,
        kwargs_str,
        ensure_hosts,
    )

    if not host.connected:
        host.connect(
            reason=f"to load fact: {name} ({kwargs_str})",
            raise_exceptions=True,
        )

    # Facts can override the shell (winrm powershell vs cmd support)
    if fact.shell_executable:
        global_kwargs["_shell_executable"] = fact.shell_executable

    command = _make_command(fact.command, fact_kwargs)
    requires_command = _make_command(fact.requires_command, fact_kwargs)
    if requires_command:
        command = StringCommand(
            # Command doesn't exist, return 0 *or* run & return fact command
            "!",
            "command",
            "-v",
            requires_command,
            ">/dev/null",
            "||",
            command,
        )

    status = False
    output = CommandOutput([])

    executor_kwargs = {
        key: value for key, value in global_kwargs.items() if key in CONNECTOR_ARGUMENT_KEYS
    }

    try:
        status, output = host.run_shell_command(
            command,
            print_output=state.print_fact_output,
            print_input=state.print_fact_input,
            **executor_kwargs,
        )
    except (timeout_error, socket_error, SSHException) as e:
        log_host_command_error(
            host,
            e,
            timeout=global_kwargs.get("_timeout"),
        )

    stdout_lines, stderr_lines = output.stdout_lines, output.stderr_lines

    data = fact.default()

    if status:
        if stdout_lines:
            try:
                data = fact.process(stdout_lines)
            except FactProcessError as e:
                log_error_or_warning(
                    host,
                    global_kwargs["_ignore_errors"],
                    description=("could not process fact: {0} {1}").format(
                        name, get_kwargs_str(fact_kwargs)
                    ),
                    exception=e,
                )

                # Check we've not failed
                if apply_failed_hosts and not global_kwargs["_ignore_errors"]:
                    state.fail_hosts({host})

    elif stderr_lines:
        # If we have error output and that error is sudo or su stating the user
        # does not exist, do not fail but instead return the default fact value.
        # This allows for users that don't currently but may be created during
        # other operations.
        first_line = stderr_lines[0]
        if executor_kwargs["_sudo_user"] and re.match(SUDO_REGEX, first_line):
            status = True
        if executor_kwargs["_su_user"] and any(re.match(regex, first_line) for regex in SU_REGEXES):
            status = True

    if status:
        log_message = "{0}{1}".format(
            host.print_prefix,
            "Loaded fact {0}{1}".format(
                click.style(name, bold=True),
                f" ({get_kwargs_str(kwargs)})" if kwargs else "",
            ),
        )
        if state.print_fact_info:
            logger.info(log_message)
        else:
            logger.debug(log_message)
    else:
        if not state.print_fact_output:
            print_host_combined_output(host, output)

        log_error_or_warning(
            host,
            global_kwargs["_ignore_errors"],
            description=("could not load fact: {0} {1}").format(name, get_kwargs_str(fact_kwargs)),
        )

    # Check we've not failed
    if apply_failed_hosts and not status and not global_kwargs["_ignore_errors"]:
        state.fail_hosts({host})

    return data



import signal
import sys

import click
import gevent

import pyinfra

from .cli import cli


def main():
    # Set CLI mode
    pyinfra.is_cli = True

    # Don't write out deploy.pyc/config.pyc etc
    sys.dont_write_bytecode = True

    sys.path.append(".")

    # Shut it click
    click.disable_unicode_literals_warning = True  # type: ignore

    # Force line buffering
    sys.stdout.reconfigure(line_buffering=True)  # type: ignore
    sys.stderr.reconfigure(line_buffering=True)  # type: ignore

    def _handle_interrupt(signum, frame):
        click.echo("Exiting upon user request!")
        sys.exit(0)

    try:
        # Kill any greenlets on ctrl+c
        gevent.signal_handler(signal.SIGINT, gevent.kill)
    except AttributeError:
        # Legacy (gevent <1.2) support
        gevent.signal(signal.SIGINT, gevent.kill)

    signal.signal(signal.SIGINT, _handle_interrupt)  # print the message and exit main
    cli()



#!/usr/bin/env python

import sys
from inspect import cleandoc, getdoc, getfullargspec
from os import makedirs, path
from typing import get_type_hints

from pyinfra.api.connectors import get_all_connectors

sys.path.append(".")
from docs.utils import title_line  # noqa: E402


def build_connectors_docs():
    this_dir = path.dirname(path.realpath(__file__))
    docs_dir = path.abspath(path.join(this_dir, "..", "docs"))

    makedirs(path.join(docs_dir, "connectors"), exist_ok=True)

    for connector_name, connector in get_all_connectors().items():
        lines = []

        full_title = "``@{0}`` Connector".format(connector_name)
        lines.append(full_title)
        lines.append(title_line("-", full_title))
        lines.append("")

        doc = getdoc(connector)
        if doc:
            lines.append(doc)
            lines.append("")

        examples_doc = getattr(connector, "__examples_doc__", None)
        if examples_doc:
            lines.append("Examples")
            lines.append(title_line("~", "Examples"))
            lines.append("")
            lines.append(cleandoc(examples_doc))
            lines.append("")
        else:
            data_title = "Usage"
            lines.append(data_title)
            lines.append(title_line("~", data_title))
            lines.append("")

            names_argument_key = getfullargspec(connector.make_names_data).args[0]
            if names_argument_key == "_":
                names_argument_key = ""
            else:
                names_argument_key = f"/{names_argument_key}"
            lines.append(
                f"""
    .. code:: shell

        pyinfra @{connector_name}{names_argument_key} ...
    """.strip(),
            )
            lines.append("")

        data_key_lines = []

        for key, type_ in get_type_hints(connector.data_cls).items():
            if key.startswith("_"):
                continue
            meta = connector.data_meta[key]
            default = "" if meta.default is None else f"``{meta.default}``"
            data_key_lines.append(
                f"""   * - ``{key}``
     - {meta.description}
     - ``{type_.__name__}``
     - {default}
"""
            )
        if data_key_lines:
            data_title = "Available Data"
            lines.append(data_title)
            lines.append(title_line("~", data_title))
            lines.append("")
            lines.append(
                "The following keys can be set as host or group data to control "
                "how this connector interacts with the target."
            )
            lines.append("")

            data_key_lines = "\n".join(data_key_lines)
            lines.append(
                f"""
.. list-table::
   :header-rows: 1
   :widths: 25 45 15 15

   * - Key
     - Description
     - Type
     - Default
{data_key_lines}
    """.strip(),
            )
            lines.append("")

        module_filename = path.join(docs_dir, "connectors", f"{connector_name}.rst")
        print("--> Writing {0}".format(module_filename))

        with open(module_filename, "w", encoding="utf-8") as outfile:
            outfile.write("\n".join(lines))


if __name__ == "__main__":
    print("### Building connectors docs")
    build_connectors_docs()



import os
from tempfile import mkstemp
from typing import TYPE_CHECKING

import click
from typing_extensions import Unpack, override

from pyinfra import logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io, memoize
from pyinfra.progress import progress_spinner

from .base import BaseConnector
from .ssh import SSHConnector
from .util import extract_control_arguments, make_unix_command_for_host

if TYPE_CHECKING:
    from pyinfra.api.arguments import ConnectorArguments
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


@memoize
def show_warning() -> None:
    logger.warning("The @dockerssh connector is in beta!")


class DockerSSHConnector(BaseConnector):
    """
    **Note**: this connector is in beta!

    The ``@dockerssh`` connector allows you to run commands on Docker containers \
    on a remote machine.

    .. code:: shell

        # A Docker base image must be provided
        pyinfra @dockerssh/remotehost:alpine:3.8 ...

        # pyinfra can run on multiple Docker images in parallel
        pyinfra @dockerssh/remotehost:alpine:3.8,@dockerssh/remotehost:ubuntu:bionic ...
    """

    handles_execution = True

    ssh: SSHConnector

    def __init__(self, state: "State", host: "Host"):
        super().__init__(state, host)
        self.ssh = SSHConnector(state, host)

    @override
    @staticmethod
    def make_names_data(name):
        try:
            hostname, image = name.split(":", 1)
        except (AttributeError, ValueError):  # failure to parse the name
            raise InventoryError("No ssh host or docker base image provided!")

        if not image:
            raise InventoryError("No docker base image provided!")

        show_warning()

        yield (
            "@dockerssh/{0}:{1}".format(hostname, image),
            {"ssh_hostname": hostname, "docker_image": image},
            ["@dockerssh"],
        )

    @override
    def connect(self) -> None:
        self.ssh.connect()

        if "docker_container_id" in self.host.host_data:  # user can provide a docker_container_id
            return

        try:
            with progress_spinner({"docker run"}):
                # last line is the container ID
                status, output = self.ssh.run_shell_command(
                    StringCommand(
                        "docker",
                        "run",
                        "-d",
                        self.host.data.docker_image,
                        "tail",
                        "-f",
                        "/dev/null",
                    ),
                )
                if not status:
                    raise IOError(output.stderr)
                container_id = output.stdout_lines[-1]

        except PyinfraError as e:
            raise ConnectError(e.args[0])

        self.host.host_data["docker_container_id"] = container_id

    @override
    def disconnect(self) -> None:
        container_id = self.host.host_data["docker_container_id"][:12]

        with progress_spinner({"docker commit"}):
            _, output = self.ssh.run_shell_command(StringCommand("docker", "commit", container_id))

            # Last line is the image ID, get sha256:[XXXXXXXXXX]...
            image_id = output.stdout_lines[-1][7:19]

        with progress_spinner({"docker rm"}):
            self.ssh.run_shell_command(
                StringCommand("docker", "rm", "-f", container_id),
            )

        logger.info(
            "{0}docker build complete, image ID: {1}".format(
                self.host.print_prefix,
                click.style(image_id, bold=True),
            ),
        )

    @override
    def run_shell_command(
        self,
        command,
        print_output: bool = False,
        print_input: bool = False,
        **arguments: Unpack["ConnectorArguments"],
    ):
        local_arguments = extract_control_arguments(arguments)

        container_id = self.host.host_data["docker_container_id"]

        command = make_unix_command_for_host(self.state, self.host, command, **arguments)
        command = QuoteString(command)

        docker_flags = "-it" if local_arguments.get("_get_pty") else "-i"
        docker_command = StringCommand(
            "docker",
            "exec",
            docker_flags,
            container_id,
            "sh",
            "-c",
            command,
        )

        return self.ssh.run_shell_command(
            docker_command,
            print_output=print_output,
            print_input=print_input,
            **local_arguments,
        )

    @override
    def put_file(
        self,
        filename_or_io,
        remote_filename,
        remote_temp_filename=None,
        print_output: bool = False,
        print_input: bool = False,
        **kwargs,  # ignored (sudo/etc)
    ):
        """
        Upload a file/IO object to the target Docker container by copying it to a
        temporary location and then uploading it into the container using ``docker cp``.
        """

        fd, local_temp_filename = mkstemp()
        remote_temp_filename = remote_temp_filename or self.host.get_temp_filename(
            local_temp_filename
        )

        # Load our file or IO object and write it to the temporary file
        with get_file_io(filename_or_io) as file_io:
            with open(local_temp_filename, "wb") as temp_f:
                data = file_io.read()

                if isinstance(data, str):
                    data = data.encode()

                temp_f.write(data)

        # upload file to remote server
        ssh_status = self.ssh.put_file(local_temp_filename, remote_temp_filename)
        if not ssh_status:
            raise IOError("Failed to copy file over ssh")

        try:
            docker_id = self.host.host_data["docker_container_id"]
            docker_command = StringCommand(
                "docker",
                "cp",
                remote_temp_filename,
                f"{docker_id}:{remote_filename}",
            )

            status, output = self.ssh.run_shell_command(
                docker_command,
                print_output=print_output,
                print_input=print_input,
            )
        finally:
            os.close(fd)
            os.remove(local_temp_filename)
            self.remote_remove(
                local_temp_filename,
                print_output=print_output,
                print_input=print_input,
            )

        if not status:
            raise IOError(output.stderr)

        if print_output:
            click.echo(
                "{0}file uploaded to container: {1}".format(
                    self.host.print_prefix,
                    remote_filename,
                ),
                err=True,
            )

        return status

    @override
    def get_file(
        self,
        remote_filename,
        filename_or_io,
        remote_temp_filename=None,
        print_output: bool = False,
        print_input: bool = False,
        **kwargs,  # ignored (sudo/etc)
    ):
        """
        Download a file from the target Docker container by copying it to a temporary
        location and then reading that into our final file/IO object.
        """

        remote_temp_filename = remote_temp_filename or self.host.get_temp_filename(remote_filename)

        try:
            docker_id = self.host.host_data["docker_container_id"]
            docker_command = StringCommand(
                "docker",
                "cp",
                f"{docker_id}:{remote_filename}",
                remote_temp_filename,
            )

            status, output = self.ssh.run_shell_command(
                docker_command,
                print_output=print_output,
                print_input=print_input,
            )

            ssh_status = self.ssh.get_file(remote_temp_filename, filename_or_io)
        finally:
            self.remote_remove(
                remote_temp_filename,
                print_output=print_output,
                print_input=print_input,
            )

        if not ssh_status:
            raise IOError("failed to copy file over ssh")

        if not status:
            raise IOError(output.stderr)

        if print_output:
            click.echo(
                "{0}file downloaded from container: {1}".format(
                    self.host.print_prefix,
                    remote_filename,
                ),
                err=True,
            )

        return status

    def remote_remove(self, filename, print_output: bool = False, print_input: bool = False):
        """
        Deletes a file on a remote machine over ssh.
        """
        remove_status, output = self.ssh.run_shell_command(
            StringCommand("rm", "-f", filename),
            print_output=print_output,
            print_input=print_input,
        )

        if not remove_status:
            raise IOError(output.stderr)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
)

from typing_extensions import TypedDict

from pyinfra.api.exceptions import ArgumentTypeError
from pyinfra.api.util import raise_if_bad_type
from pyinfra.context import ctx_config

if TYPE_CHECKING:
    from pyinfra.api import Config, Host, State

T = TypeVar("T")
default_sentinel = object()


class ArgumentMeta(Generic[T]):
    description: str
    default: Callable[["Config"], T]
    handler: Optional[Callable[["Config", T], T]]

    def __init__(self, description, default, handler=None) -> None:
        self.description = description
        self.default = default
        self.handler = handler


# Connector arguments
# These are arguments passed to the various connectors that provide the underlying
# API to read/write external systems.


# Note: ConnectorArguments is specifically not total as it's used to type many
# functions via Unpack and we don't want to specify every kwarg.
class ConnectorArguments(TypedDict, total=False):
    # Auth arguments
    _sudo: bool
    _sudo_user: str
    _use_sudo_login: bool
    _sudo_password: str
    _preserve_sudo_env: bool
    _su_user: str
    _use_su_login: bool
    _preserve_su_env: bool
    _su_shell: str
    _doas: bool
    _doas_user: str

    # Shell arguments
    _shell_executable: str
    _chdir: str
    _env: Mapping[str, str]

    # Connector control (outside of command generation)
    _success_exit_codes: Iterable[int]
    _timeout: int
    _get_pty: bool
    _stdin: Union[str, list[str], Iterable[str]]

    # Retry arguments
    _retries: int
    _retry_delay: Union[int, float]
    _retry_until: Callable[[dict], bool]

    # Temp directory argument
    _temp_dir: str


def generate_env(config: "Config", value: dict) -> dict:
    env = config.ENV.copy()
    env.update(value)
    return env


auth_argument_meta: dict[str, ArgumentMeta] = {
    "_sudo": ArgumentMeta(
        "Execute/apply any changes with sudo.",
        default=lambda config: config.SUDO,
    ),
    "_sudo_user": ArgumentMeta(
        "Execute/apply any changes with sudo as a non-root user.",
        default=lambda config: config.SUDO_USER,
    ),
    "_use_sudo_login": ArgumentMeta(
        "Execute sudo with a login shell.",
        default=lambda config: config.USE_SUDO_LOGIN,
    ),
    "_sudo_password": ArgumentMeta(
        "Password to sudo with. If needed and not specified pyinfra will prompt for it.",
        default=lambda config: config.SUDO_PASSWORD,
    ),
    "_preserve_sudo_env": ArgumentMeta(
        "Preserve the shell environment of the connecting user when using sudo.",
        default=lambda config: config.PRESERVE_SUDO_ENV,
    ),
    "_su_user": ArgumentMeta(
        "Execute/apply any changes with this user using su.",
        default=lambda config: config.SU_USER,
    ),
    "_use_su_login": ArgumentMeta(
        "Execute su with a login shell.",
        default=lambda config: config.USE_SU_LOGIN,
    ),
    "_preserve_su_env": ArgumentMeta(
        "Preserve the shell environment of the connecting user when using su.",
        default=lambda config: config.PRESERVE_SU_ENV,
    ),
    "_su_shell": ArgumentMeta(
        "Use this shell (instead of user login shell) when using ``_su``). "
        + "Only available under Linux, for use when using `su` with a user that "
        + "has nologin/similar as their login shell.",
        default=lambda config: config.SU_SHELL,
    ),
    "_doas": ArgumentMeta(
        "Execute/apply any changes with doas.",
        default=lambda config: config.DOAS,
    ),
    "_doas_user": ArgumentMeta(
        "Execute/apply any changes with doas as a non-root user.",
        default=lambda config: config.DOAS_USER,
    ),
}

shell_argument_meta: dict[str, ArgumentMeta] = {
    "_shell_executable": ArgumentMeta(
        "The shell executable to use for executing commands.",
        default=lambda config: config.SHELL,
    ),
    "_chdir": ArgumentMeta(
        "Directory to switch to before executing the command.",
        default=lambda _: None,
    ),
    "_env": ArgumentMeta(
        "Dictionary of environment variables to set.",
        default=lambda _: {},
        handler=generate_env,
    ),
    "_success_exit_codes": ArgumentMeta(
        "List of exit codes to consider a success.",
        default=lambda _: [0],
    ),
    "_timeout": ArgumentMeta(
        "Timeout for *each* command executed during the operation.",
        default=lambda _: None,
    ),
    "_get_pty": ArgumentMeta(
        "Whether to get a pseudoTTY when executing any commands.",
        default=lambda _: False,
    ),
    "_stdin": ArgumentMeta(
        "String or buffer to send to the stdin of any commands.",
        default=lambda _: None,
    ),
    "_temp_dir": ArgumentMeta(
        "Temporary directory on the remote host for file operations.",
        default=lambda config: config.TEMP_DIR,
    ),
}


# Meta arguments
# These provide/extend additional operation metadata


class MetaArguments(TypedDict):
    name: str
    _ignore_errors: bool
    _continue_on_error: bool
    _if: Union[List[Callable[[], bool]], Callable[[], bool], None]


meta_argument_meta: dict[str, ArgumentMeta] = {
    # NOTE: name is the only non-_-prefixed argument
    "name": ArgumentMeta(
        "Name of the operation.",
        default=lambda _: None,
    ),
    "_ignore_errors": ArgumentMeta(
        "Ignore errors when executing the operation.",
        default=lambda config: config.IGNORE_ERRORS,
    ),
    "_continue_on_error": ArgumentMeta(
        (
            "Continue executing operation commands after error. "
            "Only applies when ``_ignore_errors`` is true."
        ),
        default=lambda _: False,
    ),
    "_if": ArgumentMeta(
        "Only run this operation if these functions return True",
        default=lambda _: [],
    ),
}


# Execution arguments
# These alter how pyinfra is to execute an operation. Notably these must all have the same value
# over every target host for the same operation.


class ExecutionArguments(TypedDict):
    _parallel: int
    _run_once: bool
    _serial: bool


execution_argument_meta: dict[str, ArgumentMeta] = {
    "_parallel": ArgumentMeta(
        "Run this operation in batches of hosts.",
        default=lambda config: config.PARALLEL,
    ),
    "_run_once": ArgumentMeta(
        "Only execute this operation once, on the first host to see it.",
        default=lambda _: False,
    ),
    "_serial": ArgumentMeta(
        "Run this operation host by host, rather than in parallel.",
        default=lambda _: False,
    ),
}


class AllArguments(ConnectorArguments, MetaArguments, ExecutionArguments):
    pass


def all_global_arguments() -> List[tuple[str, Type]]:
    """Return all global arguments and their types."""
    return list(get_type_hints(AllArguments).items())


# Create a dictionary for retry arguments
retry_argument_meta: dict[str, ArgumentMeta] = {
    "_retries": ArgumentMeta(
        "Number of times to retry failed operations.",
        default=lambda config: config.RETRY,
    ),
    "_retry_delay": ArgumentMeta(
        "Delay in seconds between retry attempts.",
        default=lambda config: config.RETRY_DELAY,
    ),
    "_retry_until": ArgumentMeta(
        "Callable taking output data that returns True to continue retrying.",
        default=lambda config: None,
    ),
}

all_argument_meta: dict[str, ArgumentMeta] = {
    **auth_argument_meta,
    **shell_argument_meta,
    **meta_argument_meta,
    **execution_argument_meta,
    **retry_argument_meta,  # Add retry arguments
}

EXECUTION_KWARG_KEYS = list(ExecutionArguments.__annotations__.keys())
CONNECTOR_ARGUMENT_KEYS = list(ConnectorArguments.__annotations__.keys())

__argument_docs__ = {
    "Privilege & user escalation": (
        auth_argument_meta,
        """
        .. caution::
            When combining privilege escalation arguments it is important to know the order they
            are applied: ``doas`` -> ``sudo`` -> ``su``. For example
            ``_sudo=True,_su_user="pyinfra"`` yields a command like ``sudo su pyinfra..``.
        """,
        """
        .. code:: python

            # Execute a command with sudo
            server.user(
                name="Create pyinfra user using sudo",
                user="pyinfra",
                _sudo=True,
            )

            # Execute a command with a specific sudo password
            server.user(
                name="Create pyinfra user using sudo",
                user="pyinfra",
                _sudo=True,
                _sudo_password="my-secret-password",
            )
        """,
    ),
    "Shell control & features": (
        shell_argument_meta,
        "",
        """
        .. code:: python

            # Execute from a specific directory
            server.shell(
                name="Bootstrap nginx params",
                commands=["openssl dhparam -out dhparam.pem 4096"],
                _chdir="/etc/ssl/certs",
            )
        """,
    ),
    "Operation meta & callbacks": (meta_argument_meta, "", ""),
    "Execution strategy": (execution_argument_meta, "", ""),
    "Retry behavior": (
        retry_argument_meta,
        """
        Retry arguments allow you to automatically retry operations that fail. You can specify
        how many times to retry, the delay between retries, and optionally a condition
        function to determine when to stop retrying.
        """,
        """
        .. code:: python

            # Retry a command up to 3 times with the default 5 second delay
            server.shell(
                name="Run flaky command with retries",
                commands=["flaky_command"],
                _retries=3,
            )
            # Retry with a custom delay
            server.shell(
                name="Run flaky command with custom delay",
                commands=["flaky_command"],
                _retries=2,
                _retry_delay=10,  # 10 second delay between retries
            )
            # Retry with a custom condition
            def retry_on_specific_error(output_data):
                # Retry if stderr contains "temporary failure"
                for line in output_data["stderr_lines"]:
                    if "temporary failure" in line.lower():
                        return True
                return False

            server.shell(
                name="Run command with conditional retry",
                commands=["flaky_command"],
                _retries=5,
                _retry_until=retry_on_specific_error,
            )
        """,
    ),
}


def pop_global_arguments(
    state: "State",
    host: "Host",
    kwargs: dict[str, Any],
) -> tuple[AllArguments, list[str]]:
    """
    Pop and return operation global keyword arguments, in preferred order:

    + From the current context (a direct @operator or @deploy function being called)
    + From any current @deploy context (deploy kwargs)
    + From the host data variables
    + From the config variables
    """

    config = state.config
    if ctx_config.isset():
        config = ctx_config.get()
        assert config is not None

    cdkwargs = host.current_deploy_kwargs
    meta_kwargs: dict[str, Any] = cdkwargs or {}  # type: ignore[assignment]

    arguments: dict[str, Any] = {}
    found_keys: list[str] = []

    for key, type_ in all_global_arguments():
        argument_meta = all_argument_meta[key]
        handler = argument_meta.handler
        default: Any = argument_meta.default(config)

        host_default = getattr(host.data, key, default_sentinel)
        if host_default is not default_sentinel:
            default = host_default

        if key in kwargs:
            found_keys.append(key)
            value = kwargs.pop(key)
        else:
            value = meta_kwargs.get(key, default)

        if handler:
            value = handler(config, value)

        if value != default:
            raise_if_bad_type(
                value,
                type_,
                ArgumentTypeError,
                f"Invalid argument `{key}`:",
            )

        # TODO: why is type failing here?
        arguments[key] = value  # type: ignore
    return cast(AllArguments, arguments), found_keys



#!/usr/bin/env python

import sys
from inspect import cleandoc
from os import path
from typing import get_type_hints

from pyinfra.api import Config
from pyinfra.api.arguments import AllArguments, __argument_docs__
from pyinfra.api.host import Host
from pyinfra.api.operation import OperationMeta

sys.path.append(".")
from docs.utils import title_line  # noqa: E402


def build_arguments_doc():
    pyinfra_config = Config()

    this_dir = path.dirname(path.realpath(__file__))
    docs_dir = path.abspath(path.join(this_dir, "..", "docs"))

    lines = []

    # Extend locals with hidden (behind TYPE_CHECKING) imports in the arguments module
    locals_ = locals()
    locals_["Host"] = Host
    locals_["OperationMeta"] = OperationMeta

    all_arguments = get_type_hints(AllArguments)

    for group_name, (
        arguments_meta,
        arguments_title_doc,
        arguments_example_doc,
    ) in __argument_docs__.items():
        lines.append("\n{0}".format(group_name))
        lines.append(title_line("~", group_name))
        lines.append("")

        if arguments_title_doc:
            lines.append(cleandoc(arguments_title_doc))
            lines.append("")

        lines.append(
            """.. list-table::
   :header-rows: 1
   :widths: 25 45 15 15

   * - Key
     - Description
     - Type
     - Default"""
        )

        for key, meta in arguments_meta.items():
            default = meta.default
            if callable(default):
                default = default(pyinfra_config)
            default = "" if default is None else f"``{default}``"

            type_ = all_arguments[key]
            type_name = type_.__name__
            if hasattr(type_, "__args__"):
                type_args = ", ".join([arg.__name__ for arg in type_.__args__])
                type_name = f"{type_name}[{type_args}]"

            lines.append(
                f"""   * - ``{key}``
     - {meta.description}
     - ``{type_name}``
     - {default}
"""
            )

        if arguments_example_doc:
            lines.append("**Examples:**")
            lines.append("")
            lines.append(cleandoc(arguments_example_doc))

    module_filename = path.join(docs_dir, "_deploy_globals.rst")
    print("--> Writing {0}".format(module_filename))

    out = "\n".join(lines)

    with open(module_filename, "w", encoding="utf-8") as outfile:
        outfile.write(out)


if __name__ == "__main__":
    print("### Building arguments doc")
    build_arguments_doc()



import re
from typing import Dict, List, Optional, TypedDict, Union

from typing_extensions import NotRequired, override

from pyinfra.api import FactBase
from pyinfra.api.util import try_int


class CrontabDict(TypedDict):
    command: NotRequired[str]
    # handles cases like CRON_TZ=UTC
    env: NotRequired[str]
    minute: NotRequired[Union[int, str]]
    hour: NotRequired[Union[int, str]]
    month: NotRequired[Union[int, str]]
    day_of_month: NotRequired[Union[int, str]]
    day_of_week: NotRequired[Union[int, str]]
    comments: NotRequired[List[str]]
    special_time: NotRequired[str]


# for compatibility, also keeps a dict of command -> crontab dict
class CrontabFile:
    commands: List[CrontabDict]

    def __init__(self, input_dict: Optional[Dict[str, CrontabDict]] = None):
        super().__init__()
        self.commands = []
        if input_dict:
            for command, others in input_dict.items():
                val = others.copy()
                val["command"] = command
                self.add_item(val)

    def add_item(self, item: CrontabDict):
        self.commands.append(item)

    def __len__(self):
        return len(self.commands)

    def __bool__(self):
        return len(self) > 0

    def items(self):
        return {item.get("command") or item.get("env"): item for item in self.commands}

    def get_command(
        self, command: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[CrontabDict]:
        assert command or name, "Either command or name must be provided"

        name_comment = "# pyinfra-name={0}".format(name)
        for cmd in self.commands:
            if "command" not in cmd:
                continue
            if cmd.get("command") == command:
                return cmd
            if name_comment in cmd.get("comments", []):
                return cmd
        return None

    def get_env(self, env: str) -> Optional[CrontabDict]:
        for cmd in self.commands:
            if cmd.get("env") == env:
                return cmd
        return None

    def get(self, item: str) -> Optional[CrontabDict]:
        return self.get_command(command=item, name=item) or self.get_env(item)

    def __getitem__(self, item) -> Optional[CrontabDict]:
        return self.get(item)

    @override
    def __repr__(self):
        return f"CrontabResult({self.commands})"

    # noinspection PyMethodMayBeStatic
    def format_item(self, item: CrontabDict):
        lines = []
        for comment in item.get("comments", []):
            lines.append(comment)

        if "env" in item:
            lines.append(item["env"])
        elif "special_time" in item:
            lines.append(f"{item['special_time']} {item['command']}")
        else:
            lines.append(
                f"{item['minute']} {item['hour']} "
                f"{item['day_of_month']} {item['month']} {item['day_of_week']} "
                f"{item['command']}"
            )
        return "\n".join(lines)

    @override
    def __str__(self):
        return "\n".join(self.format_item(item) for item in self.commands)

    def to_json(self):
        return self.commands


_crontab_env_re = re.compile(r"^\s*([A-Z_]+)=(.*)$")


class Crontab(FactBase[CrontabFile]):
    """
    Returns a dictionary of CrontabFile.

    .. code:: python

        # CrontabFile.items()
        {
            "/path/to/command": {
                "minute": "*",
                "hour": "*",
                "month": "*",
                "day_of_month": "*",
                "day_of_week": "*",
            },
            "echo another command": {
                "special_time": "@daily",
            },
        }
        # or CrontabFile.to_json()
        [
            {
                "command": "/path/to/command",
                "minute": "*",
                "hour": "*",
                "month": "*",
                "day_of_month": "*",
                "day_of_week": "*",
            },
            {
                "command": "echo another command",
                "special_time": "@daily",
            }
        ]
    """

    default = CrontabFile

    @override
    def requires_command(self, user=None) -> str:
        return "crontab"

    @override
    def command(self, user=None):
        if user:
            return "crontab -l -u {0} || true".format(user)
        return "crontab -l || true"

    @override
    def process(self, output):
        crons = CrontabFile()
        current_comments = []

        for line in output:
            line = line.strip()
            if not line or line.startswith("#"):
                current_comments.append(line)
                continue

            if line.startswith("@"):
                special_time, command = line.split(None, 1)
                item = CrontabDict(
                    command=command,
                    special_time=special_time,
                    comments=current_comments,
                )
                crons.add_item(item)

            elif _crontab_env_re.match(line):
                # handle environment variables
                item = CrontabDict(
                    env=line,
                    comments=current_comments,
                )
                crons.add_item(item)
            else:
                minute, hour, day_of_month, month, day_of_week, command = line.split(None, 5)
                item = CrontabDict(
                    command=command,
                    minute=try_int(minute),
                    hour=try_int(hour),
                    month=try_int(month),
                    day_of_month=try_int(day_of_month),
                    day_of_week=try_int(day_of_week),
                    comments=current_comments,
                )
                crons.add_item(item)

            current_comments = []
        return crons



"""
Manage flatpak packages. See https://www.flatpak.org/
"""

from __future__ import annotations

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.flatpak import FlatpakPackages


@operation()
def packages(
    packages: str | list[str] | None = None,
    remote: str | None = None,
    present=True,
):
    """
    Install/remove a flatpak package

    + packages: List of packages
    + remote: Source to install the application or runtime from
    + present: whether the package should be installed

    **Examples:**

    .. code:: python

        from pyinfra.operations import flatpak
        # Install vlc flatpak
        flatpak.package(
            name="Install vlc",
            packages="org.videolan.VLC",
        )

        # Install vlc flatpak from flathub
        flatpak.package(
            name="Install vlc",
            packages="org.videolan.VLC",
            remote="flathub",
        )

        # Install multiple flatpaks
        flatpak.package(
            name="Install vlc and kodi",
            packages=["org.videolan.VLC", "tv.kodi.Kodi"],
        )

        # Remove vlc
        flatpak.package(
            name="Remove vlc",
            packages="org.videolan.VLC",
            present=False,
        )
    """

    if packages is None:
        return

    if isinstance(packages, str):
        packages = [packages]

    flatpak_packages = host.get_fact(FlatpakPackages)

    install_packages = []
    remove_packages = []

    if remote is None:
        remote = ""
    else:
        # ensure we have a space between the remote and packages
        remote = remote.strip() + " "

    for package in packages:
        # it's installed
        if package in flatpak_packages:
            if not present:
                # we don't want it
                remove_packages.append(package)

        # it's not installed
        if package not in flatpak_packages:
            # we want it
            if present:
                install_packages.append(package)

            # we don't want it
            else:
                host.noop(f"flatpak package {package} is not installed")

    if install_packages:
        yield f"flatpak install --noninteractive {remote}{' '.join(install_packages)}"

    if remove_packages:
        yield f"flatpak uninstall --noninteractive {' '.join(remove_packages)}"



"""
Manage packages on OpenWrt using opkg
    + ``update`` - update local copy of package information
    + ``packages`` -  install and remove packages

see https://openwrt.org/docs/guide-user/additional-software/opkg
OpenWrt recommends against upgrading all packages  thus there is no ``opkg.upgrade`` function
"""

from typing import List, Union

from pyinfra import host
from pyinfra.api import StringCommand, operation
from pyinfra.facts.opkg import OpkgPackages
from pyinfra.operations.util.packaging import ensure_packages

EQUALS = "="


@operation(is_idempotent=False)
def update():
    """
    Update the local opkg information.
    """

    yield StringCommand("opkg update")


_update = update


@operation()
def packages(
    packages: Union[str, List[str]] = "",
    present: bool = True,
    latest: bool = False,
    update: bool = True,
):
    """
    Add/remove/update opkg packages.

    + packages: package or list of packages to that must/must not be present
    + present: whether the package(s) should be installed (default True) or removed
    + latest: whether to attempt to upgrade the specified package(s) (default False)
    + update: run ``opkg update`` before installing packages (default True)

    Not Supported:
        Opkg does not support version pinning, i.e. ``<pkg>=<version>`` is not allowed
        and will cause an exception.

    **Examples:**

    .. code:: python

        from pyinfra.operations import opkg
        # Ensure packages are installed (will not force package upgrade)
        opkg.packages(['asterisk', 'vim'], name="Install Asterisk and Vim")

        # Install the latest versions of packages (always check)
        opkg.packages(
            'vim',
            latest=True,
            name="Ensure we have the latest version of Vim"
        )
    """
    if str(packages) == "" or (
        isinstance(packages, list) and (len(packages) < 1 or all(len(p) < 1 for p in packages))
    ):
        host.noop("empty or invalid package list provided to opkg.packages")
        return

    pkg_list = packages if isinstance(packages, list) else [packages]
    have_equals = ",".join([pkg.split(EQUALS)[0] for pkg in pkg_list if EQUALS in pkg])
    if len(have_equals) > 0:
        raise ValueError(f"opkg does not support version pinning but found for: '{have_equals}'")

    if update:
        yield from _update._inner()

    yield from ensure_packages(
        host,
        pkg_list,
        host.get_fact(OpkgPackages),
        present,
        install_command="opkg install",
        upgrade_command="opkg upgrade",
        uninstall_command="opkg remove",
        latest=latest,
    )



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from enum import IntEnum
from graphlib import CycleError, TopologicalSorter
from multiprocessing import cpu_count
from typing import TYPE_CHECKING, Callable, Iterator, Optional

from gevent.pool import Pool
from paramiko import PKey

from pyinfra import logger

from .config import Config
from .exceptions import PyinfraError

if TYPE_CHECKING:
    from pyinfra.api.arguments import AllArguments
    from pyinfra.api.command import PyinfraCommand
    from pyinfra.api.host import Host
    from pyinfra.api.inventory import Inventory
    from pyinfra.api.operation import OperationMeta


# Work out the max parallel we can achieve with the open files limit of the user/process,
# take 10 for opening Python files and /3 for ~3 files per host during op runs.
# See: https://github.com/Fizzadar/pyinfra/issues/44
try:
    from resource import RLIMIT_NOFILE, getrlimit

    nofile_limit, _ = getrlimit(RLIMIT_NOFILE)
    MAX_PARALLEL = round((nofile_limit - 10) / 3)

# Resource isn't available on Windows
except ImportError:
    nofile_limit = 0
    MAX_PARALLEL = 100000


class BaseStateCallback:
    # Host callbacks
    #

    @staticmethod
    def host_before_connect(state: "State", host: "Host"):
        pass

    @staticmethod
    def host_connect(state: "State", host: "Host"):
        pass

    @staticmethod
    def host_connect_error(state: "State", host: "Host", error):
        pass

    @staticmethod
    def host_disconnect(state: "State", host: "Host"):
        pass

    # Operation callbacks
    #

    @staticmethod
    def operation_start(state: "State", op_hash):
        pass

    @staticmethod
    def operation_host_start(state: "State", host: "Host", op_hash):
        pass

    @staticmethod
    def operation_host_success(state: "State", host: "Host", op_hash, retry_count: int = 0):
        pass

    @staticmethod
    def operation_host_error(
        state: "State", host: "Host", op_hash, retry_count: int = 0, max_retries: int = 0
    ):
        pass

    @staticmethod
    def operation_host_retry(
        state: "State", host: "Host", op_hash, retry_num: int, max_retries: int
    ):
        pass

    @staticmethod
    def operation_end(state: "State", op_hash):
        pass


class StateStage(IntEnum):
    # Setup - collect inventory & data
    Setup = 1
    # Connect - connect to the inventory
    Connect = 2
    # Prepare - detect operation changes
    Prepare = 3
    # Execute - execute operations
    Execute = 4
    # Disconnect - disconnect from the inventory
    Disconnect = 5


class StateOperationMeta:
    names: set[str]
    args: list[str]
    op_order: tuple[int, ...]
    global_arguments: "AllArguments"

    def __init__(self, op_order: tuple[int, ...]):
        self.op_order = op_order
        self.names = set()
        self.args = []
        self.global_arguments = {}  # type: ignore


@dataclass
class StateOperationHostData:
    command_generator: Callable[[], Iterator["PyinfraCommand"]]
    global_arguments: "AllArguments"
    operation_meta: "OperationMeta"
    parent_op_hash: Optional[str] = None


class StateHostMeta:
    ops = 0
    ops_change = 0
    ops_no_change = 0
    op_hashes: set[str]

    def __init__(self) -> None:
        self.op_hashes = set()


class StateHostResults:
    ops = 0
    success_ops = 0
    error_ops = 0
    ignored_error_ops = 0
    partial_ops = 0


class State:
    """
    Manages state for a pyinfra deploy.
    """

    initialised: bool = False

    # A pyinfra.api.Inventory which stores all our pyinfra.api.Host's
    inventory: "Inventory"

    # A pyinfra.api.Config
    config: "Config"

    # Main gevent pool
    pool: "Pool"

    # Current stage this state is in
    current_stage: StateStage = StateStage.Setup
    # Warning counters by stage
    stage_warnings: dict[StateStage, int] = defaultdict(int)

    # Whether we are executing operations (ie hosts are all ready)
    is_executing: bool = False

    # Whether we should check for operation changes as part of the operation ordering phase, this
    # allows us to guesstimate which ops will result in changes on which hosts.
    check_for_changes: bool = True

    print_noop_info: bool = False  # print "[host] noop: reason for noop"
    print_fact_info: bool = False  # print "loaded fact X"
    print_input: bool = False
    print_fact_input: bool = False
    print_output: bool = False
    print_fact_output: bool = False

    # Used in CLI
    cwd: Optional[str] = None  # base directory for locating files/templates/etc
    current_deploy_filename: Optional[str] = None
    current_exec_filename: Optional[str] = None
    current_op_file_number: int = 0
    should_raise_failed_hosts: Optional[Callable[["State"], bool]] = None

    def __init__(
        self,
        inventory: Optional["Inventory"] = None,
        config: Optional["Config"] = None,
        check_for_changes: bool = True,
        **kwargs,
    ):
        """
        Initializes the state, the main Pyinfra

        Args:
            inventory (Optional[Inventory], optional): The inventory. Defaults to None.
            config (Optional[Config], optional): The config object. Defaults to None.
        """
        self.check_for_changes = check_for_changes

        if inventory:
            self.init(inventory, config, **kwargs)

    def init(
        self,
        inventory: "Inventory",
        config: Optional["Config"],
        initial_limit=None,
    ):
        # Config validation
        #

        # If no config, create one using the defaults
        if config is None:
            config = Config()

        if not config.PARALLEL:
            # TODO: benchmark this
            # In my own tests the optimum number of parallel SSH processes is
            # ~20 per CPU core - no science here yet, needs benchmarking!
            cpus = cpu_count()
            ideal_parallel = cpus * 20

            config.PARALLEL = min(ideal_parallel, len(inventory), MAX_PARALLEL)

        # If explicitly set, just issue a warning
        elif config.PARALLEL > MAX_PARALLEL:
            logger.warning(
                (
                    "Parallel set to {0}, but this may hit the open files limit of {1}.\n"
                    "    Max recommended value: {2}"
                ).format(config.PARALLEL, nofile_limit, MAX_PARALLEL),
            )

        # Actually initialise the state object
        #

        self.callback_handlers: list[BaseStateCallback] = []

        # Setup greenlet pools
        self.pool = Pool(config.PARALLEL)
        self.fact_pool = Pool(config.PARALLEL)

        # Private keys
        self.private_keys: dict[str, PKey] = {}

        # Assign inventory/config
        self.inventory = inventory
        self.config = config

        # Hosts we've activated at any time
        self.activated_hosts: set["Host"] = set()
        # Active hosts that *haven't* failed yet
        self.active_hosts: set["Host"] = set()
        # Hosts that have failed
        self.failed_hosts: set["Host"] = set()

        # Limit hosts changes dynamically to limit operations to a subset of hosts
        self.limit_hosts: list["Host"] = initial_limit

        # Op basics
        self.op_meta: dict[str, StateOperationMeta] = {}  # maps operation hash -> names/etc

        # Op dict for each host
        self.ops: dict["Host", dict[str, StateOperationHostData]] = {host: {} for host in inventory}

        # Meta dict for each host
        self.meta: dict["Host", StateHostMeta] = {host: StateHostMeta() for host in inventory}

        # Results dict for each host
        self.results: dict["Host", StateHostResults] = {
            host: StateHostResults() for host in inventory
        }

        # Assign state back references to inventory & config
        inventory.state = config.state = self
        for host in inventory:
            host.init(self)

        self.initialised = True

    def set_stage(self, stage: StateStage) -> None:
        if stage < self.current_stage:
            raise Exception("State stage cannot go backwards!")
        self.current_stage = stage

    def increment_warning_counter(self) -> None:
        self.stage_warnings[self.current_stage] += 1

    def get_warning_counter(self) -> int:
        return self.stage_warnings[self.current_stage]

    def should_check_for_changes(self):
        return self.check_for_changes

    def add_callback_handler(self, handler):
        if not isinstance(handler, BaseStateCallback):
            raise TypeError(
                ("{0} is not a valid callback handler (use `BaseStateCallback`)").format(handler),
            )
        self.callback_handlers.append(handler)

    def trigger_callbacks(self, method_name: str, *args, **kwargs):
        for handler in self.callback_handlers:
            func = getattr(handler, method_name)
            func(self, *args, **kwargs)

    def get_op_order(self):
        ts: TopologicalSorter = TopologicalSorter()

        for host in self.inventory:
            for i, op_hash in enumerate(host.op_hash_order):
                if not i:
                    ts.add(op_hash)
                else:
                    ts.add(op_hash, host.op_hash_order[i - 1])

        final_op_order = []

        try:
            ts.prepare()
        except CycleError as e:
            raise PyinfraError(
                (
                    "Cycle detected in operation ordering DAG.\n"
                    f"    Error: {e}\n\n"
                    "    This can happen when using loops in operation code, "
                    "please see: https://docs.pyinfra.com/en/latest/deploy-process.html#loops-cycle-errors"  # noqa: E501
                ),
            )

        while ts.is_active():
            # Ensure that where we have multiple different operations that can be executed in any
            # dependency order we order them by line numbers.
            node_group = sorted(
                ts.get_ready(),
                key=lambda op_hash: self.op_meta[op_hash].op_order,
            )
            ts.done(*node_group)
            final_op_order.extend(node_group)

        return final_op_order

    def get_op_meta(self, op_hash: str) -> StateOperationMeta:
        return self.op_meta[op_hash]

    def get_meta_for_host(self, host: "Host") -> StateHostMeta:
        return self.meta[host]

    def get_results_for_host(self, host: "Host") -> StateHostResults:
        return self.results[host]

    def get_op_data_for_host(
        self,
        host: "Host",
        op_hash: str,
    ) -> StateOperationHostData:
        return self.ops[host][op_hash]

    def set_op_data_for_host(
        self,
        host: "Host",
        op_hash: str,
        op_data: StateOperationHostData,
    ):
        self.ops[host][op_hash] = op_data

    def activate_host(self, host: "Host"):
        """
        Flag a host as active.
        """

        logger.debug("Activating host: %s", host)

        # Add to *both* activated and active - active will reduce as hosts fail
        # but connected will not, enabling us to track failed %.
        self.activated_hosts.add(host)
        self.active_hosts.add(host)

    def fail_hosts(self, hosts_to_fail, activated_count=None):
        """
        Flag a ``set`` of hosts as failed, error for ``config.FAIL_PERCENT``.
        """

        if not hosts_to_fail:
            return

        activated_count = activated_count or len(self.activated_hosts)

        logger.debug(
            "Failing hosts: {0}".format(
                ", ".join(
                    (host.name for host in hosts_to_fail),
                ),
            ),
        )

        self.failed_hosts.update(hosts_to_fail)

        self.active_hosts -= hosts_to_fail

        # Check we're not above the fail percent
        active_hosts = self.active_hosts

        # No hosts left!
        if not active_hosts:
            raise PyinfraError("No hosts remaining!")

        if self.config.FAIL_PERCENT is not None:
            percent_failed = (1 - len(active_hosts) / activated_count) * 100

            if percent_failed > self.config.FAIL_PERCENT:
                if self.should_raise_failed_hosts and self.should_raise_failed_hosts(self) is False:
                    return

                raise PyinfraError(
                    "Over {0}% of hosts failed ({1}%)".format(
                        self.config.FAIL_PERCENT,
                        int(round(percent_failed)),
                    ),
                )

    def is_host_in_limit(self, host: "Host"):
        """
        Returns a boolean indicating if the host is within the current state limit.
        """

        limit_hosts = self.limit_hosts

        if not isinstance(limit_hosts, list):
            return True
        return host in limit_hosts



import re
import subprocess
from typing import Callable

import pytest
import testinfra.host


class Helpers:
    @staticmethod
    def run(command, cwd=None, expected_exit_code=0):
        results = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=cwd,
        )

        stdout, stderr = results.communicate()

        if isinstance(stdout, bytes):
            stdout = stdout.decode("utf-8")
        if isinstance(stderr, bytes):
            stderr = stderr.decode("utf-8")

        assert results.returncode == expected_exit_code, stderr

        return stdout, stderr

    @staticmethod
    def run_check_output(command, expected_lines=None, **kwargs):
        if expected_lines is None:
            expected_lines = ["Connected", "Starting operation", "Errors: 0"]

        _, stderr = Helpers.run(command, **kwargs)

        for line in expected_lines:
            assert re.search(line, stderr, re.MULTILINE), 'Line "{0}" not found in output!'.format(
                line,
            )

    @staticmethod
    def run_container_test_host(
        container: str,
        cmd: str,
        callback: Callable[[testinfra.host.Host], None],
    ) -> None:
        stdout, _ = Helpers.run(
            f"docker run -d {container} tail -f /dev/null",
        )
        cid = stdout.strip()

        try:
            Helpers.run_check_output(
                f"pyinfra -y @docker/{cid} {cmd}",
                expected_lines=["docker build complete"],
            )

            host = testinfra.get_host(f"docker://{cid}")
            callback(host)
        finally:
            Helpers.run(f"docker rm -f {cid}")


@pytest.fixture(scope="module")
def helpers():
    return Helpers



import json
from os import path
from queue import Queue
from threading import Thread

from typing_extensions import override

from pyinfra import local, logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import memoize
from pyinfra.progress import progress_spinner

from .base import BaseConnector


def _get_vagrant_ssh_config(queue, progress, target):
    logger.debug("Loading SSH config for %s", target)

    queue.put(
        local.shell(
            "vagrant ssh-config {0}".format(target),
            splitlines=True,
        ),
    )

    progress(target)


@memoize
def get_vagrant_config(limit=None):
    logger.info("Getting Vagrant config...")

    if limit and not isinstance(limit, (list, tuple)):
        limit = [limit]

    with progress_spinner({"vagrant status"}) as progress:
        output = local.shell(
            "vagrant status --machine-readable",
            splitlines=True,
        )
        progress("vagrant status")

    targets = []

    for line in output:
        line = line.strip()
        _, target, type_, data = line.split(",", 3)

        # Skip anything not in the limit
        if limit is not None and target not in limit:
            continue

        # For each running container - fetch it's SSH config in a thread - this
        # is because Vagrant *really* slow to run each command.
        if type_ == "state" and data == "running":
            targets.append(target)

    threads = []
    config_queue = Queue()  # type: ignore

    with progress_spinner(targets) as progress:
        for target in targets:
            thread = Thread(
                target=_get_vagrant_ssh_config,
                args=(config_queue, progress, target),
            )
            threads.append(thread)
            thread.start()

    for thread in threads:
        thread.join()

    queue_items = list(config_queue.queue)

    lines = []
    for output in queue_items:
        lines.extend([ln.strip() for ln in output])

    return lines


@memoize
def get_vagrant_options():
    if path.exists("@vagrant.json"):
        with open("@vagrant.json", "r", encoding="utf-8") as f:
            return json.loads(f.read())
    return {}


def _make_name_data(host):
    vagrant_options = get_vagrant_options()
    vagrant_host = host["Host"]

    data = {
        "ssh_hostname": host["HostName"],
    }

    for config_key, data_key, data_cast in (
        ("Port", "ssh_port", int),
        ("User", "ssh_user", str),
        ("IdentityFile", "ssh_key", str),
    ):
        if config_key in host:
            data[data_key] = data_cast(host[config_key])

    # Update any configured JSON data
    if vagrant_host in vagrant_options.get("data", {}):
        data.update(vagrant_options["data"][vagrant_host])

    # Work out groups
    groups = vagrant_options.get("groups", {}).get(vagrant_host, [])

    if "@vagrant" not in groups:
        groups.append("@vagrant")

    return "@vagrant/{0}".format(host["Host"]), data, groups


class VagrantInventoryConnector(BaseConnector):
    """
    The ``@vagrant`` connector reads the current Vagrant status and generates an
    inventory for any running VMs.

    .. code:: shell

        # Run on all hosts
        pyinfra @vagrant ...

        # Run on a specific VM
        pyinfra @vagrant/my-vm-name ...

        # Run on multiple named VMs
        pyinfra @vagrant/my-vm-name,@vagrant/another-vm-name ...
    """

    @override
    @staticmethod
    def make_names_data(name=None):
        vagrant_ssh_info = get_vagrant_config(name)

        logger.debug("Got Vagrant SSH info: \n%s", vagrant_ssh_info)

        hosts = []
        current_host = None

        for line in vagrant_ssh_info:
            # Vagrant outputs an empty line between each host
            if not line:
                if current_host:
                    hosts.append(_make_name_data(current_host))

                current_host = None
                continue

            key, value = line.split(" ", 1)

            if key == "Host":
                if current_host:
                    hosts.append(_make_name_data(current_host))

                # Set the new host
                current_host = {
                    key: value,
                }

            elif current_host:
                current_host[key] = value

            else:
                logger.debug("Extra Vagrant SSH key/value (%s=%s)", key, value)

        if current_host:
            hosts.append(_make_name_data(current_host))

        if not hosts:
            if name:
                raise InventoryError(
                    "No running Vagrant instances matching `{0}` found!".format(name)
                )
            raise InventoryError("No running Vagrant instances found!")

        for host in hosts:
            yield host



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generator, Generic, Iterable, List, Mapping, Union

from typing_extensions import ParamSpec, Protocol

if TYPE_CHECKING:
    from pyinfra.api.operation import OperationMeta

P = ParamSpec("P")


# Unfortunately we have to re-type out all of the global arguments here because
# Python typing doesn't (yet) support merging kwargs. This acts as the operation
# decorator output function which merges actual operation args/kwargs in paramspec
# with the global arguments available in all operations.
# The nature of "global arguments" is somewhat opposed to static typing as it's
# indirect and somewhat magic, but we are where we are.
class PyinfraOperation(Generic[P], Protocol):
    _inner: Callable[P, Generator]

    def __call__(
        self,
        #
        # ConnectorArguments
        #
        # Auth
        _sudo: bool = False,
        _sudo_user: None | str = None,
        _use_sudo_login: bool = False,
        _sudo_password: None | str = None,
        _preserve_sudo_env: bool = False,
        _su_user: None | str = None,
        _use_su_login: bool = False,
        _preserve_su_env: bool = False,
        _su_shell: None | str = None,
        _doas: bool = False,
        _doas_user: None | str = None,
        # Shell arguments
        _shell_executable: None | str = None,
        _chdir: None | str = None,
        _env: None | Mapping[str, str] = None,
        # Connector control
        _success_exit_codes: Iterable[int] = (0,),
        _timeout: None | int = None,
        _get_pty: bool = False,
        _stdin: None | Union[str, list[str], Iterable[str]] = None,
        # Retry arguments
        _retries: None | int = None,
        _retry_delay: None | Union[int, float] = None,
        _retry_until: None | Callable[[dict], bool] = None,
        _temp_dir: None | str = None,
        #
        # MetaArguments
        #
        name: None | str = None,
        _ignore_errors: bool = False,
        _continue_on_error: bool = False,
        _if: Union[List[Callable[[], bool]], Callable[[], bool], None] = None,
        #
        # ExecutionArguments
        #
        _parallel: None | int = None,
        _run_once: bool = False,
        _serial: bool = False,
        #
        # op args
        #
        *args: P.args,
        #
        # op kwargs
        #
        **kwargs: P.kwargs,
    ) -> "OperationMeta": ...



from __future__ import annotations

from typing_extensions import override

from pyinfra.api import FactBase

from .util.packaging import parse_packages

# Source: https://superuser.com/a/1472405
# Modified to return version and release inside a single group and removed extra capturing groups
APK_REGEX = r"(.+)-([^-]+-r[^-]+) \S+ \{\S+\} \(.+?\)"


class ApkPackages(FactBase):
    """
    Returns a dict of installed apk packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    @override
    def command(self) -> str:
        return "apk list --installed"

    @override
    def requires_command(self) -> str:
        return "apk"

    default = dict

    @override
    def process(self, output):
        return parse_packages(APK_REGEX, output)



from os import path

from pyinfra import inventory
from pyinfra.context import ctx_inventory, ctx_state

from ..paramiko_util import PatchSSHTestCase
from .util import run_cli


class TestCliInventory(PatchSSHTestCase):
    def test_load_deploy_group_data(self):
        ctx_state.reset()
        ctx_inventory.reset()

        hosts = ["somehost", "anotherhost", "someotherhost"]
        result = run_cli(
            "-y",
            ",".join(hosts),
            path.join("tests", "test_cli", "deploy", "deploy.py"),
            f"--chdir={path.join('tests', 'test_cli', 'deploy')}",
        )
        assert result.exit_code == 0, result.stdout

        assert inventory.data.get("hello") == "world"
        assert "leftover_data" in inventory.group_data
        assert inventory.group_data["leftover_data"].get("still_parsed") == "never_used"
        assert inventory.group_data["leftover_data"].get("_global_arg") == "gets_parsed"

    def test_load_group_data(self):
        ctx_state.reset()
        ctx_inventory.reset()

        hosts = ["somehost", "anotherhost", "someotherhost"]
        result = run_cli(
            "-y",
            ",".join(hosts),
            f"--group-data={path.join('tests', 'test_cli', 'deploy', 'group_data')}",
            "exec",
            "uptime",
        )
        assert result.exit_code == 0, result.stdout

        assert inventory.data.get("hello") == "world"
        assert "leftover_data" in inventory.group_data
        assert inventory.group_data["leftover_data"].get("still_parsed") == "never_used"
        assert inventory.group_data["leftover_data"].get("_global_arg") == "gets_parsed"

    def test_load_group_data_file(self):
        ctx_state.reset()
        ctx_inventory.reset()

        hosts = ["somehost", "anotherhost", "someotherhost"]
        filename = path.join("tests", "test_cli", "deploy", "group_data", "leftover_data.py")
        result = run_cli(
            "-y",
            ",".join(hosts),
            f"--group-data={filename}",
            "exec",
            "uptime",
        )
        assert result.exit_code == 0, result.stdout

        assert "hello" not in inventory.data
        assert "leftover_data" in inventory.group_data
        assert inventory.group_data["leftover_data"].get("still_parsed") == "never_used"
        assert inventory.group_data["leftover_data"].get("_global_arg") == "gets_parsed"

    def test_ignores_variables_with_leading_underscore(self):
        ctx_state.reset()
        ctx_inventory.reset()

        result = run_cli(
            path.join("tests", "test_cli", "inventories", "invalid.py"),
            "exec",
            "--debug",
            "--",
            "echo hi",
        )

        assert result.exit_code == 0, result.stdout
        assert (
            'Ignoring variable "_hosts" in inventory file since it starts with a leading underscore'
            in result.stderr
        )
        assert inventory.hosts == {}

    def test_only_supports_list_and_tuples(self):
        ctx_state.reset()
        ctx_inventory.reset()

        result = run_cli(
            path.join("tests", "test_cli", "inventories", "invalid.py"),
            "exec",
            "--debug",
            "--",
            "echo hi",
        )

        assert result.exit_code == 0, result.stdout
        assert 'Ignoring variable "dict_hosts" in inventory file' in result.stderr, result.stdout
        assert 'Ignoring variable "generator_hosts" in inventory file' in result.stderr, (
            result.stdout
        )
        assert inventory.hosts == {}

    def test_host_groups_may_only_contain_strings_or_tuples(self):
        ctx_state.reset()
        ctx_inventory.reset()

        result = run_cli(
            path.join("tests", "test_cli", "inventories", "invalid.py"),
            "exec",
            "--",
            "echo hi",
        )

        assert result.exit_code == 0, result.stdout
        assert 'Ignoring host group "issue_662"' in result.stderr, result.stdout
        assert inventory.hosts == {}



# Variables names with a leading underscore are ignored
_hosts = ["host-1", "host-2", "host-3"]

# Dictionaries are not supported
dict_hosts = {
    "host-1": "hostname",
}

# Generators are not supported
generator_hosts = (host for host in ["host-1", "host-2"])

# https://github.com/pyinfra-dev/pyinfra/issues/662
issue_662 = ["a", "b", ({"foo": "bar"})]



"""
The files operations handles filesystem state, file uploads and template generation.
"""

from __future__ import annotations

import os
import posixpath
import sys
import traceback
from datetime import datetime, timedelta, timezone
from fnmatch import fnmatch
from io import StringIO
from pathlib import Path
from typing import IO, Any, Union

import click
from jinja2 import TemplateRuntimeError, TemplateSyntaxError, UndefinedError

from pyinfra import host, logger, state
from pyinfra.api import (
    FileDownloadCommand,
    FileUploadCommand,
    OperationError,
    OperationTypeError,
    OperationValueError,
    QuoteString,
    RsyncCommand,
    StringCommand,
    operation,
)
from pyinfra.api.command import make_formatted_string_command
from pyinfra.api.util import (
    get_call_location,
    get_file_io,
    get_file_md5,
    get_file_sha1,
    get_file_sha256,
    get_path_permissions_mode,
    get_template,
    memoize,
)
from pyinfra.facts.files import (
    MARKER_BEGIN_DEFAULT,
    MARKER_DEFAULT,
    MARKER_END_DEFAULT,
    Block,
    Directory,
    File,
    FileContents,
    FindFiles,
    FindInFile,
    Flags,
    Link,
    Md5File,
    Sha1File,
    Sha256File,
    Sha384File,
)
from pyinfra.facts.server import Date, Which

from .util import files as file_utils
from .util.files import (
    MetadataTimeField,
    adjust_regex,
    ensure_mode_int,
    generate_color_diff,
    get_timestamp,
    sed_delete,
    sed_replace,
    unix_path_join,
)


@operation()
def download(
    src: str,
    dest: str,
    user: str | None = None,
    group: str | None = None,
    mode: str | None = None,
    cache_time: int | None = None,
    force=False,
    sha384sum: str | None = None,
    sha256sum: str | None = None,
    sha1sum: str | None = None,
    md5sum: str | None = None,
    headers: dict[str, str] | None = None,
    insecure=False,
    proxy: str | None = None,
    temp_dir: str | Path | None = None,
    extra_curl_args: dict[str, str] | None = None,
    extra_wget_args: dict[str, str] | None = None,
):
    """
    Download files from remote locations using ``curl`` or ``wget``.

    + src: source URL of the file
    + dest: where to save the file
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + cache_time: if the file exists already, re-download after this time (in seconds)
    + force: always download the file, even if it already exists
    + sha384sum: sha384 hash to checksum the downloaded file against
    + sha256sum: sha256 hash to checksum the downloaded file against
    + sha1sum: sha1 hash to checksum the downloaded file against
    + md5sum: md5 hash to checksum the downloaded file against
    + headers: optional dictionary of headers to set for the HTTP request
    + insecure: disable SSL verification for the HTTP request
    + proxy: simple HTTP proxy through which we can download files, form `http://<yourproxy>:<port>`
    + temp_dir: use this custom temporary directory during the download
    + extra_curl_args: optional dictionary with custom arguments for curl
    + extra_wget_args: optional dictionary with custom arguments for wget

    **Example:**

    .. code:: python

        from pyinfra.operations import files
        files.download(
            name="Download the Docker repo file",
            src="https://download.docker.com/linux/centos/docker-ce.repo",
            dest="/etc/yum.repos.d/docker-ce.repo",
        )
    """

    info = host.get_fact(File, path=dest)

    # Destination is a directory?
    if info is False:
        raise OperationError(
            "Destination {0} already exists and is not a file".format(dest),
        )

    # Do we download the file? Force by default
    download = force

    # Doesn't exist, lets download it
    if info is None:
        download = True

    # Destination file exists & cache_time: check when the file was last modified,
    # download if old
    else:
        if cache_time:
            # Time on files is not tz-aware, and will be the same tz as the server's time,
            # so we can safely remove the tzinfo from the Date fact before comparison.
            ctime = host.get_fact(Date).replace(tzinfo=None) - timedelta(seconds=cache_time)
            if info["mtime"] and info["mtime"] < ctime:
                download = True

        if sha1sum:
            if sha1sum != host.get_fact(Sha1File, path=dest):
                download = True

        if sha256sum:
            if sha256sum != host.get_fact(Sha256File, path=dest):
                download = True

        if sha384sum:
            if sha384sum != host.get_fact(Sha384File, path=dest):
                download = True

        if md5sum:
            if md5sum != host.get_fact(Md5File, path=dest):
                download = True

    # If we download, always do user/group/mode as SSH user may be different
    if download:
        # Use explicit temp_dir if provided, otherwise fall back to _temp_dir global argument
        effective_temp_dir = temp_dir
        if effective_temp_dir is None and host.current_op_global_arguments:
            effective_temp_dir = host.current_op_global_arguments.get("_temp_dir")
        temp_file = host.get_temp_filename(
            dest, temp_directory=str(effective_temp_dir) if effective_temp_dir is not None else None
        )

        curl_args: list[Union[str, StringCommand]] = ["-sSLf"]
        wget_args: list[Union[str, StringCommand]] = ["-q"]

        if extra_curl_args:
            for key, value in extra_curl_args.items():
                curl_args.append(StringCommand(key, QuoteString(value)))

        if extra_wget_args:
            for key, value in extra_wget_args.items():
                wget_args.append(StringCommand(key, QuoteString(value)))

        if proxy:
            curl_args.append(f"--proxy {proxy}")
            wget_args.append("-e use_proxy=yes")
            wget_args.append(f"-e http_proxy={proxy}")

        if insecure:
            curl_args.append("--insecure")
            wget_args.append("--no-check-certificate")

        if headers:
            for key, value in headers.items():
                header_arg = StringCommand("--header", QuoteString(f"{key}: {value}"))
                curl_args.append(header_arg)
                wget_args.append(header_arg)

        curl_command = make_formatted_string_command(
            "curl {0} {1} -o {2}",
            StringCommand(*curl_args),
            QuoteString(src),
            QuoteString(temp_file),
        )
        wget_command = make_formatted_string_command(
            "wget {0} {1} -O {2} || ( rm -f {2} ; exit 1 )",
            StringCommand(*wget_args),
            QuoteString(src),
            QuoteString(temp_file),
        )

        if host.get_fact(Which, command="curl"):
            yield curl_command
        elif host.get_fact(Which, command="wget"):
            yield wget_command
        else:
            yield "( {0} ) || ( {1} )".format(curl_command, wget_command)

        yield StringCommand("mv", QuoteString(temp_file), QuoteString(dest))

        if user or group:
            yield file_utils.chown(dest, user, group)

        if mode:
            yield file_utils.chmod(dest, mode)

        if sha1sum:
            yield make_formatted_string_command(
                (
                    "(( sha1sum {0} 2> /dev/null || shasum {0} || sha1 {0} ) | grep {1} ) "
                    "|| ( echo {2} && exit 1 )"
                ),
                QuoteString(dest),
                sha1sum,
                QuoteString("SHA1 did not match!"),
            )

        if sha256sum:
            yield make_formatted_string_command(
                (
                    "(( sha256sum {0} 2> /dev/null || shasum -a 256 {0} || sha256 {0} ) "
                    "| grep {1}) || ( echo {2} && exit 1 )"
                ),
                QuoteString(dest),
                sha256sum,
                QuoteString("SHA256 did not match!"),
            )

        if sha384sum:
            yield make_formatted_string_command(
                (
                    "(( sha384sum {0} 2> /dev/null || shasum -a 384 {0} ) "
                    "| grep {1}) || ( echo {2} && exit 1 )"
                ),
                QuoteString(dest),
                sha384sum,
                QuoteString("SHA384 did not match!"),
            )

        if md5sum:
            yield make_formatted_string_command(
                ("(( md5sum {0} 2> /dev/null || md5 {0} ) | grep {1}) || ( echo {2} && exit 1 )"),
                QuoteString(dest),
                md5sum,
                QuoteString("MD5 did not match!"),
            )
    else:
        host.noop("file {0} has already been downloaded".format(dest))


@operation()
def line(
    path: str,
    line: str,
    present=True,
    replace: str | None = None,
    flags: list[str] | None = None,
    backup=False,
    interpolate_variables=False,
    escape_regex_characters=False,
    ensure_newline=False,
):
    """
    Ensure lines in files using grep to locate and sed to replace.

    + path: target remote file to edit
    + line: string or regex matching the target line
    + present: whether the line should be in the file
    + replace: text to replace entire matching lines when ``present=True``
    + flags: list of flags to pass to sed when replacing/deleting
    + backup: whether to backup the file (see below)
    + interpolate_variables: whether to interpolate variables in ``replace``
    + escape_regex_characters: whether to escape regex characters from the matching line
    + ensure_newline: ensures that the appended line is on a new line

    Regex line matching:
        Unless line matches a line (starts with ^, ends $), pyinfra will wrap it such that
        it does, like: ``^.*LINE.*$``. This means we don't swap parts of lines out. To
        change bits of lines, see ``files.replace``.

    Regex line escaping:
        If matching special characters (eg a crontab line containing ``*``), remember to escape
        it first using Python's ``re.escape``.

    Backup:
        If set to ``True``, any editing of the file will place an old copy with the ISO
        date (taken from the machine running ``pyinfra``) appended as the extension. If
        you pass a string value this will be used as the extension of the backed up file.

    Append:
        If ``line`` is not in the file but we want it (``present`` set to ``True``), then
        it will be append to the end of the file.

    Ensure new line:
        This will ensure that the ``line`` being appended is always on a separate new
        line in case the file doesn't end with a newline character.


    **Examples:**

    .. code:: python

        # prepare to do some maintenance
        maintenance_line = "SYSTEM IS DOWN FOR MAINTENANCE"
        files.line(
            name="Add the down-for-maintenance line in /etc/motd",
            path="/etc/motd",
            line=maintenance_line,
        )

        # Then, after the maintenance is done, remove the maintenance line
        files.line(
            name="Remove the down-for-maintenance line in /etc/motd",
            path="/etc/motd",
            line=maintenance_line,
            replace="",
            present=False,
        )

        # example where there is '*' in the line
        files.line(
            name="Ensure /netboot/nfs is in /etc/exports",
            path="/etc/exports",
            line=r"/netboot/nfs .*",
            replace="/netboot/nfs *(ro,sync,no_wdelay,insecure_locks,no_root_squash,"
            "insecure,no_subtree_check)",
        )

        files.line(
            name="Ensure myweb can run /usr/bin/python3 without password",
            path="/etc/sudoers",
            line=r"myweb .*",
            replace="myweb ALL=(ALL) NOPASSWD: /usr/bin/python3",
        )

        # example when there are double quotes (")
        line = 'QUOTAUSER=""'
        files.line(
            name="Example with double quotes (")",
            path="/etc/adduser.conf",
            line="^{}$".format(line),
            replace=line,
        )
    """

    match_line = adjust_regex(line, escape_regex_characters)
    #
    # if escape_regex_characters:
    #     match_line = re.sub(r"([\.\\\+\*\?\[\^\]\$\(\)\{\}\-])", r"\\\1", match_line)
    #
    # # Ensure we're matching a whole line, note: match may be a partial line so we
    # # put any matches on either side.
    # if not match_line.startswith("^"):
    #     match_line = "^.*{0}".format(match_line)
    # if not match_line.endswith("$"):
    #     match_line = "{0}.*$".format(match_line)

    # Is there a matching line in this file?
    present_lines = host.get_fact(
        FindInFile,
        path=path,
        pattern=match_line,
        interpolate_variables=interpolate_variables,
    )

    # If replace present, use that over the matching line
    if replace:
        line = replace
    # We must provide some kind of replace to sed_replace_command below
    else:
        replace = ""

    # Save commands for re-use in dynamic script when file not present at fact stage
    if ensure_newline:
        echo_command = make_formatted_string_command(
            "( [ $(tail -c1 {1} | wc -l) -eq 0 ] && echo ; echo {0} ) >> {1}",
            '"{0}"'.format(line) if interpolate_variables else QuoteString(line),
            QuoteString(path),
        )
    else:
        echo_command = make_formatted_string_command(
            "echo {0} >> {1}",
            '"{0}"'.format(line) if interpolate_variables else QuoteString(line),
            QuoteString(path),
        )

    if backup:
        backup_filename = "{0}.{1}".format(path, get_timestamp())
        echo_command = StringCommand(
            make_formatted_string_command(
                "cp {0} {1} && ",
                QuoteString(path),
                QuoteString(backup_filename),
            ),
            echo_command,
        )

    sed_replace_command = sed_replace(
        path,
        match_line,
        replace,
        flags=flags,
        backup=backup,
        interpolate_variables=interpolate_variables,
    )

    # No line and we want it, append it
    if not present_lines and present:
        # If we're doing replacement, only append if the *replacement* line
        # does not exist (as we are appending the replacement).
        if replace:
            # Ensure replace explicitly matches a whole line
            replace_line = replace
            if not replace_line.startswith("^"):
                replace_line = f"^{replace_line}"
            if not replace_line.endswith("$"):
                replace_line = f"{replace_line}$"

            present_lines = host.get_fact(
                FindInFile,
                path=path,
                pattern=replace_line,
                interpolate_variables=interpolate_variables,
            )

        if not present_lines:
            yield echo_command
        else:
            host.noop('line "{0}" exists in {1}'.format(replace or line, path))

    # Line(s) exists and we want to remove them
    elif present_lines and not present:
        yield sed_delete(
            path,
            match_line,
            "",
            flags=flags,
            backup=backup,
            interpolate_variables=interpolate_variables,
        )

    # Line(s) exists and we have want to ensure they're correct
    elif present_lines and present:
        # If any of lines are different, sed replace them
        if replace and any(line != replace for line in present_lines):
            yield sed_replace_command
        else:
            host.noop('line "{0}" exists in {1}'.format(replace or line, path))


@operation()
def replace(
    path: str,
    text: str | None = None,
    replace: str | None = None,
    flags: list[str] | None = None,
    backup=False,
    interpolate_variables=False,
    match=None,  # deprecated
):
    """
    Replace contents of a file using ``sed``.

    + path: target remote file to edit
    + text: text/regex to match against
    + replace: text to replace with
    + flags: list of flags to pass to sed
    + backup: whether to backup the file (see below)
    + interpolate_variables: whether to interpolate variables in ``replace``

    Backup:
        If set to ``True``, any editing of the file will place an old copy with the ISO
        date (taken from the machine running ``pyinfra``) appended as the extension. If
        you pass a string value this will be used as the extension of the backed up file.

    **Example:**

    .. code:: python

        files.replace(
            name="Change part of a line in a file",
            path="/etc/motd",
            text="verboten",
            replace="forbidden",
        )
    """

    if text is None and match:
        text = match
        logger.warning(
            (
                "The `match` argument has been replaced by "
                "`text` in the `files.replace` operation ({0})"
            ).format(get_call_location()),
        )

    if text is None:
        raise TypeError("Missing argument `text` required in `files.replace` operation")

    if replace is None:
        raise TypeError("Missing argument `replace` required in `files.replace` operation")

    existing_lines = host.get_fact(
        FindInFile,
        path=path,
        pattern=text,
        interpolate_variables=interpolate_variables,
    )

    # Only do the replacement if the file does not exist (it may be created earlier)
    # or we have matching lines.
    if existing_lines is None or existing_lines:
        yield sed_replace(
            path,
            text,
            replace,
            flags=flags,
            backup=backup,
            interpolate_variables=interpolate_variables,
        )
    else:
        host.noop('string "{0}" does not exist in {1}'.format(text, path))


@operation()
def sync(
    src: str,
    dest: str,
    user: str | None = None,
    group: str | None = None,
    mode: str | None = None,
    dir_mode: str | None = None,
    delete=False,
    exclude: str | list[str] | tuple[str] | None = None,
    exclude_dir: str | list[str] | tuple[str] | None = None,
    add_deploy_dir=True,
):
    """
    Syncs a local directory with a remote one, with delete support. Note that delete will
    remove extra files on the remote side, but not extra directories.

    + src: local directory to sync
    + dest: remote directory to sync to
    + user: user to own the files and directories
    + group: group to own the files and directories
    + mode: permissions of the files
    + dir_mode: permissions of the directories
    + delete: delete remote files not present locally
    + exclude: string or list/tuple of strings to match & exclude files (eg ``*.pyc``)
    + exclude_dir: string or list/tuple of strings to match & exclude directories (eg node_modules)
    + add_deploy_dir: interpret src as relative to deploy directory instead of current directory

    **Example:**

    .. code:: python

        # Sync local files/tempdir to remote /tmp/tempdir
        files.sync(
            name="Sync a local directory with remote",
            src="files/tempdir",
            dest="/tmp/tempdir",
        )

    Note: ``exclude`` and ``exclude_dir`` use ``fnmatch`` behind the scenes to do the filtering.

    + ``exclude`` matches against the filename.
    + ``exclude_dir`` matches against the path of the directory, relative to ``src``.
      Since fnmatch does not treat path separators (``/`` or ``\\``) as special characters,
      excluding all directories matching a given name, however deep under ``src`` they are,
      can be done for example with ``exclude_dir=["__pycache__", "*/__pycache__"]``

    """
    original_src = src  # Keep a copy to reference in errors
    src = os.path.normpath(src)

    # Add deploy directory?
    if add_deploy_dir and state.cwd:
        src = os.path.join(state.cwd, src)

    # Ensure the source directory exists
    if not os.path.isdir(src):
        raise IOError("No such directory: {0}".format(original_src))

    # Ensure exclude is a list/tuple
    if exclude is not None:
        if not isinstance(exclude, (list, tuple)):
            exclude = [exclude]

    # Ensure exclude_dir is a list/tuple
    if exclude_dir is not None:
        if not isinstance(exclude_dir, (list, tuple)):
            exclude_dir = [exclude_dir]

    put_files = []
    ensure_dirnames = []
    for dirpath, dirnames, filenames in os.walk(src, topdown=True):
        remote_dirpath = Path(os.path.normpath(os.path.relpath(dirpath, src))).as_posix()

        # Filter excluded dirs
        for child_dir in dirnames[:]:
            child_path = os.path.normpath(os.path.join(remote_dirpath, child_dir))
            if exclude_dir and any(fnmatch(child_path, match) for match in exclude_dir):
                dirnames.remove(child_dir)

        if remote_dirpath and remote_dirpath != os.path.curdir:
            ensure_dirnames.append((remote_dirpath, get_path_permissions_mode(dirpath)))

        for filename in filenames:
            full_filename = os.path.join(dirpath, filename)

            # Should we exclude this file?
            if exclude and any(fnmatch(full_filename, match) for match in exclude):
                continue

            remote_full_filename = unix_path_join(
                *[
                    item
                    for item in (dest, remote_dirpath, filename)
                    if item and item != os.path.curdir
                ]
            )
            put_files.append((full_filename, remote_full_filename))

    # Ensure the destination directory - if the destination is a link, ensure
    # the link target is a directory.
    dest_to_ensure = dest
    dest_link_info = host.get_fact(Link, path=dest)
    if dest_link_info:
        dest_to_ensure = dest_link_info["link_target"]

    yield from directory._inner(
        path=dest_to_ensure,
        user=user,
        group=group,
        mode=dir_mode or get_path_permissions_mode(src),
    )

    # Ensure any remote dirnames
    for dir_path_curr, dir_mode_curr in ensure_dirnames:
        yield from directory._inner(
            path=unix_path_join(dest, dir_path_curr),
            user=user,
            group=group,
            mode=dir_mode or dir_mode_curr,
        )

    # Put each file combination
    for local_filename, remote_filename in put_files:
        yield from put._inner(
            src=local_filename,
            dest=remote_filename,
            user=user,
            group=group,
            mode=mode or get_path_permissions_mode(local_filename),
            add_deploy_dir=False,
            create_remote_dir=False,  # handled above
        )

    # Delete any extra files
    if delete:
        remote_filenames = set(host.get_fact(FindFiles, path=dest) or [])
        wanted_filenames = set([remote_filename for _, remote_filename in put_files])
        files_to_delete = remote_filenames - wanted_filenames
        for filename in files_to_delete:
            # Should we exclude this file?
            if exclude and any(fnmatch(filename, match) for match in exclude):
                continue

            yield from file._inner(path=filename, present=False)


@memoize
def show_rsync_warning() -> None:
    logger.warning("The `files.rsync` operation is in alpha!")


@operation(is_idempotent=False)
def rsync(src: str, dest: str, flags: list[str] | None = None):
    """
    Use ``rsync`` to sync a local directory to the remote system. This operation will actually call
    the ``rsync`` binary on your system.

    .. important::
        The ``files.rsync`` operation is in alpha, and only supported using SSH
        or ``@local`` connectors. When using the SSH connector, rsync will automatically use the
        StrictHostKeyChecking setting, config and known_hosts file (when specified).

    .. caution::
        When using SSH, the ``files.rsync`` operation only supports the ``sudo`` and ``sudo_user``
        global arguments.
    """

    if flags is None:
        flags = ["-ax", "--delete"]
    show_rsync_warning()

    try:
        host.check_can_rsync()
    except NotImplementedError as e:
        raise OperationError(*e.args)

    yield RsyncCommand(src, dest, flags)


def _create_remote_dir(remote_filename, user, group):
    # Always use POSIX style path as local might be Windows, remote always *nix
    remote_dirname = posixpath.dirname(remote_filename)
    if remote_dirname:
        yield from directory._inner(
            path=remote_dirname,
            user=user,
            group=group,
            _no_check_owner_mode=True,  # don't check existing user/mode
            _no_fail_on_link=True,  # don't fail if the path is a link
        )


def _file_equal(local_path: str | IO[Any] | None, remote_path: str) -> bool:
    if local_path is None:
        return False
    for fact, get_sum in [
        (Sha1File, get_file_sha1),
        (Md5File, get_file_md5),
        (Sha256File, get_file_sha256),
    ]:
        remote_sum = host.get_fact(fact, path=remote_path)
        if remote_sum:
            local_sum = get_sum(local_path)
            return local_sum == remote_sum
    return False


def _remote_file_equal(remote_path_a: str, remote_path_b: str) -> bool:
    for fact in [Sha1File, Md5File, Sha256File]:
        sum_a = host.get_fact(fact, path=remote_path_a)
        sum_b = host.get_fact(fact, path=remote_path_b)
        if sum_a and sum_b:
            return sum_a == sum_b
    return False


@operation(
    # We don't (currently) cache the local state, so there's nothing we can
    # update to flag the local file as present.
    is_idempotent=False,
)
def get(
    src: str,
    dest: str,
    add_deploy_dir=True,
    create_local_dir=False,
    force=False,
):
    """
    Download a file from the remote system.

    + src: the remote filename to download
    + dest: the local filename to download the file to
    + add_deploy_dir: dest is relative to the deploy directory
    + create_local_dir: create the local directory if it doesn't exist
    + force: always download the file, even if the local copy matches

    Note:
        This operation is not suitable for large files as it may involve copying
        the remote file before downloading it.

    **Example:**

    .. code:: python

        files.get(
            name="Download a file from a remote",
            src="/etc/centos-release",
            dest="/tmp/whocares",
        )
    """

    if add_deploy_dir and state.cwd:
        dest = os.path.join(state.cwd, dest)

    if create_local_dir:
        local_pathname = os.path.dirname(dest)
        if not os.path.exists(local_pathname):
            os.makedirs(local_pathname)

    remote_file = host.get_fact(File, path=src)

    # Use _temp_dir global argument if provided
    temp_dir = None
    if host.current_op_global_arguments:
        temp_dir = host.current_op_global_arguments.get("_temp_dir")

    # No remote file, so assume exists and download it "blind"
    if not remote_file or force:
        yield FileDownloadCommand(
            src, dest, remote_temp_filename=host.get_temp_filename(dest, temp_directory=temp_dir)
        )

    # No local file, so always download
    elif not os.path.exists(dest):
        yield FileDownloadCommand(
            src, dest, remote_temp_filename=host.get_temp_filename(dest, temp_directory=temp_dir)
        )

    # Remote file exists - check if it matches our local
    else:
        # Check hash sum, download if needed
        if not _file_equal(dest, src):
            yield FileDownloadCommand(
                src,
                dest,
                remote_temp_filename=host.get_temp_filename(dest, temp_directory=temp_dir),
            )
        else:
            host.noop("file {0} has already been downloaded".format(dest))


def _canonicalize_timespec(field: MetadataTimeField, local_file, timespec):
    if isinstance(timespec, datetime):
        if not timespec.tzinfo:
            # specify remote host timezone
            timespec_with_tz = timespec.replace(tzinfo=host.get_fact(Date).tzinfo)
            return timespec_with_tz
        else:
            return timespec
    elif isinstance(timespec, bool) and timespec:
        lf_ts = (
            os.stat(local_file).st_atime
            if field is MetadataTimeField.ATIME
            else os.stat(local_file).st_mtime
        )
        return datetime.fromtimestamp(lf_ts, tz=timezone.utc)
    else:
        try:
            isodatetime = datetime.fromisoformat(timespec)
            if not isodatetime.tzinfo:
                return isodatetime.replace(tzinfo=host.get_fact(Date).tzinfo)
            else:
                return isodatetime
        except ValueError:
            try:
                timestamp = float(timespec)
                return datetime.fromtimestamp(timestamp, tz=timezone.utc)
            except ValueError:
                # verify there is a remote file matching path in timesrc
                ref_file = host.get_fact(File, path=timespec)
                if ref_file:
                    if field is MetadataTimeField.ATIME:
                        assert ref_file["atime"] is not None
                        return ref_file["atime"].replace(tzinfo=timezone.utc)
                    else:
                        assert ref_file["mtime"] is not None
                        return ref_file["mtime"].replace(tzinfo=timezone.utc)
                else:
                    ValueError("Bad argument for `timesspec`: {0}".format(timespec))


# returns True for a visible difference in the second field between the datetime values
#   in the ref's TZ
def _times_differ_in_s(ref, cand):
    assert ref.tzinfo and cand.tzinfo
    cand_in_ref_tz = cand.astimezone(ref.tzinfo)
    return (abs((cand_in_ref_tz - ref).total_seconds()) >= 1.0) or (
        ref.second != cand_in_ref_tz.second
    )


@operation()
def put(
    src: str | IO[Any],
    dest: str,
    user: str | None = None,
    group: str | None = None,
    mode: int | str | bool | None = None,
    add_deploy_dir=True,
    create_remote_dir=True,
    force=False,
    assume_exists=False,
    atime: datetime | float | int | str | bool | None = None,
    mtime: datetime | float | int | str | bool | None = None,
):
    """
    Upload a local file, or file-like object, to the remote system.

    + src: filename or IO-like object to upload
    + dest: remote filename to upload to
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files, use ``True`` to copy the local file
    + add_deploy_dir: src is relative to the deploy directory
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: always upload the file, even if the remote copy matches
    + assume_exists: whether to assume the local file exists
    + atime: value of atime the file should have, use ``True`` to match the local file
    + mtime: value of mtime the file should have, use ``True`` to match the local file

    ``dest``:
        If this is a directory that already exists on the remote side, the local
        file will be uploaded to that directory with the same filename.

    ``mode``:
        When set to ``True`` the permissions of the local file are applied to the
        remote file after the upload is complete.  If set to an octal value with
        digits for at least user, group, and other, either as an ``int`` or
        ``str``, those permissions will be used.

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    ``atime`` and ``mtime``:
        When set to values other than ``False`` or ``None``, the respective metadata
        fields on the remote file will updated accordingly.  Timestamp values are
        considered equivalent if the difference is less than one second and they have
        the identical number in the seconds field.  If set to ``True`` the local
        file is the source of the value.  Otherwise, these values can be provided as
        ``datetime`` objects, POSIX timestamps, or strings that can be parsed into
        either of these date and time specifications.  They can also be reference file
        paths on the remote host, as with the ``-r`` argument to ``touch``.  If a
        ``datetime`` argument has no ``tzinfo`` value (i.e., it is naive), it is
        assumed to be in the remote host's local timezone.  There is no shortcut for
        setting both ``atime` and ``mtime`` values with a single time specification,
        unlike the native ``touch`` command.

    Notes:
        This operation is not suitable for large files as it may involve copying
        the file before uploading it.

        Currently, if the mode argument is anything other than a ``bool`` or a full
        octal permission set and the remote file exists, the operation will always
        behave as if the remote file does not match the specified permissions and
        requires a change.

        If the ``atime`` argument is set for a given file, unless the remote
        filesystem is mounted ``noatime`` or ``relatime``, multiple runs of this
        operation will trigger the change detection for that file, since the act of
        reading and checksumming the file will cause the host OS to update the file's
        ``atime``.

    **Examples:**

    .. code:: python

        files.put(
            name="Update the message of the day file",
            src="files/motd",
            dest="/etc/motd",
            mode="644",
        )

        files.put(
            name="Upload a StringIO object",
            src=StringIO("file contents"),
            dest="/etc/motd",
        )
    """

    # Upload IO objects as-is
    if hasattr(src, "read"):
        local_file = src
        local_sum_path = src

    # Assume string filename
    else:
        assert isinstance(src, (str, Path))
        # Add deploy directory?
        if add_deploy_dir and state.cwd:
            src = os.path.join(state.cwd, src)

        local_file = src

        if os.path.isfile(local_file):
            local_sum_path = local_file
        elif assume_exists:
            local_sum_path = None
        else:
            raise IOError("No such file: {0}".format(local_file))

    if mode is True:
        if isinstance(local_file, str) and os.path.isfile(local_file):
            mode = get_path_permissions_mode(local_file)
        else:
            logger.warning(
                ("No local file exists to get permissions from with `mode=True` ({0})").format(
                    get_call_location(),
                ),
            )
    else:
        mode = ensure_mode_int(mode)

    remote_file = host.get_fact(File, path=dest)

    if not remote_file and bool(host.get_fact(Directory, path=dest)):
        assert isinstance(src, str)
        dest = unix_path_join(dest, os.path.basename(src))
        remote_file = host.get_fact(File, path=dest)

    if create_remote_dir:
        yield from _create_remote_dir(dest, user, group)

    # Use _temp_dir global argument if provided
    temp_dir = None
    if host.current_op_global_arguments:
        temp_dir = host.current_op_global_arguments.get("_temp_dir")

    # No remote file, always upload and user/group/mode if supplied
    if not remote_file or force:
        if state.config.DIFF:
            host.log(f"Will create {click.style(dest, bold=True)}", logger.info)

            with get_file_io(src, "r") as f:
                desired_lines = f.readlines()

            for line in generate_color_diff([], desired_lines):
                logger.info(f"  {line}")
            logger.info("")

        yield FileUploadCommand(
            local_file,
            dest,
            remote_temp_filename=host.get_temp_filename(dest, temp_directory=temp_dir),
        )

        if user or group:
            yield file_utils.chown(dest, user, group)

        if mode:
            yield file_utils.chmod(dest, mode)

        # do mtime before atime to ensure atime setting isn't undone by mtime setting
        if mtime:
            yield file_utils.touch(
                dest,
                MetadataTimeField.MTIME,
                _canonicalize_timespec(MetadataTimeField.MTIME, src, mtime),
            )

        if atime:
            yield file_utils.touch(
                dest,
                MetadataTimeField.ATIME,
                _canonicalize_timespec(MetadataTimeField.ATIME, src, atime),
            )

    # File exists, check sum and check user/group/mode/atime/mtime if supplied
    else:
        if not _file_equal(local_sum_path, dest):
            if state.config.DIFF:
                # Generate diff when contents change
                current_contents = host.get_fact(FileContents, path=dest)
                if current_contents:
                    current_lines = [line + "\n" for line in current_contents]
                else:
                    current_lines = []

                host.log(f"Will modify {click.style(dest, bold=True)}", logger.info)

                with get_file_io(src, "r") as f:
                    desired_lines = f.readlines()

                for line in generate_color_diff(current_lines, desired_lines):
                    logger.info(f"  {line}")
                logger.info("")

            yield FileUploadCommand(
                local_file,
                dest,
                remote_temp_filename=host.get_temp_filename(dest, temp_directory=temp_dir),
            )

            if user or group:
                yield file_utils.chown(dest, user, group)

            if mode:
                yield file_utils.chmod(dest, mode)

            if mtime:
                yield file_utils.touch(
                    dest,
                    MetadataTimeField.MTIME,
                    _canonicalize_timespec(MetadataTimeField.MTIME, src, mtime),
                )

            if atime:
                yield file_utils.touch(
                    dest,
                    MetadataTimeField.ATIME,
                    _canonicalize_timespec(MetadataTimeField.ATIME, src, atime),
                )

        else:
            changed = False

            # Check mode
            if mode and remote_file["mode"] != mode:
                yield file_utils.chmod(dest, mode)
                changed = True

            # Check user/group
            if (user and remote_file["user"] != user) or (group and remote_file["group"] != group):
                yield file_utils.chown(dest, user, group)
                changed = True

            # Check mtime
            if mtime:
                canonical_mtime = _canonicalize_timespec(MetadataTimeField.MTIME, src, mtime)
                assert remote_file["mtime"] is not None
                if _times_differ_in_s(
                    canonical_mtime, remote_file["mtime"].replace(tzinfo=timezone.utc)
                ):
                    yield file_utils.touch(dest, MetadataTimeField.MTIME, canonical_mtime)
                    changed = True

            # Check atime
            if atime:
                canonical_atime = _canonicalize_timespec(MetadataTimeField.ATIME, src, atime)
                assert remote_file["atime"] is not None
                if _times_differ_in_s(
                    canonical_atime, remote_file["atime"].replace(tzinfo=timezone.utc)
                ):
                    yield file_utils.touch(dest, MetadataTimeField.ATIME, canonical_atime)
                    changed = True

            if not changed:
                host.noop("file {0} is already uploaded".format(dest))


@operation()
def template(
    src: str | IO[Any],
    dest: str,
    user: str | None = None,
    group: str | None = None,
    mode: str | None = None,
    create_remote_dir: bool = True,
    jinja_env_kwargs: dict[str, Any] | None = None,
    **data,
):
    '''
    Generate a template using jinja2 and write it to the remote system.

    + src: template filename or IO-like object
    + dest: remote filename
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + create_remote_dir: create the remote directory if it doesn't exist
    + jinja_env_kwargs: keyword arguments to be passed into the jinja Environment()

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    ``jinja_env_kwargs``:
        To have more control over how jinja2 renders your template, you can pass
        a dict with arguments that will be passed as keyword args to the jinja2
        `Environment() <https://jinja.palletsprojects.com/en/3.0.x/api/#jinja2.Environment>`_.

    The ``host``, ``state``, and ``inventory`` objects will be automatically passed to the template
    if not set explicitly.

    Notes:
        Common convention is to store templates in a "templates" directory and
        have a filename suffix with '.j2' (for jinja2).

        The default template lookup directory (used with jinjas ``extends``, ``import`` and
        ``include`` statements) is the current working directory.

        For information on the template syntax, see
        `the jinja2 docs <https://jinja.palletsprojects.com>`_.

    **Examples:**

    .. code:: python

        files.template(
            name="Create a templated file",
            src="templates/somefile.conf.j2",
            dest="/etc/somefile.conf",
        )

        files.template(
            name="Create service file",
            src="templates/myweb.service.j2",
            dest="/etc/systemd/system/myweb.service",
            mode="755",
            user="root",
            group="root",
        )

        # Example showing how to pass python variable to template file. You can also
        # use dicts and lists. The .j2 file can use `{{ foo_variable }}` to be interpolated.
        foo_variable = 'This is some foo variable contents'
        foo_dict = {
            "str1": "This is string 1",
            "str2": "This is string 2"
        }
        foo_list = [
            "entry 1",
            "entry 2"
        ]

        template = StringIO("""
        name: "{{ foo_variable }}"
        dict_contents:
            str1: "{{ foo_dict.str1 }}"
            str2: "{{ foo_dict.str2 }}"
        list_contents:
        {% for entry in foo_list %}
            - "{{ entry }}"
        {% endfor %}
        """)

        files.template(
            name="Create a templated file",
            src=template,
            dest="/tmp/foo.yml",
            foo_variable=foo_variable,
            foo_dict=foo_dict,
            foo_list=foo_list
        )

        # Example showing how to use host and inventory in a template file.
        template = StringIO("""
        name: "{{ host.name }}"
        list_contents:
        {% for entry in inventory.groups.my_servers %}
            - "{{ entry }}"
        {% endfor %}
        """)

        files.template(
            name="Create a templated file",
            src=template,
            dest="/tmp/foo.yml"
        )
    '''

    if not hasattr(src, "read") and state.cwd:
        src = os.path.join(state.cwd, src)

    # Ensure host/state/inventory are available inside templates (if not set)
    data.setdefault("host", host)
    data.setdefault("state", state)
    data.setdefault("inventory", state.inventory)

    # Render and make file-like it's output
    try:
        output = get_template(src, jinja_env_kwargs).render(data)
    except (TemplateRuntimeError, TemplateSyntaxError, UndefinedError) as e:
        trace_frames = [
            frame
            for frame in traceback.extract_tb(sys.exc_info()[2])
            if frame[2] in ("template", "<module>", "top-level template code")
        ]  # thank you https://github.com/saltstack/salt/blob/master/salt/utils/templates.py

        line_number = trace_frames[-1][1]

        # Quickly read the line in question and one above/below for nicer debugging
        with get_file_io(src, "r") as f:
            template_lines = f.readlines()

        template_lines = [line.strip() for line in template_lines]
        relevant_lines = template_lines[max(line_number - 2, 0) : line_number + 1]

        raise OperationError(
            "Error in template: {0} (L{1}): {2}\n...\n{3}\n...".format(
                src,
                line_number,
                e,
                "\n".join(relevant_lines),
            ),
        ) from None

    output_file = StringIO(output)
    # Set the template attribute for nicer debugging
    output_file.template = src  # type: ignore[attr-defined]

    # Pass to the put function
    yield from put._inner(
        src=output_file,
        dest=dest,
        user=user,
        group=group,
        mode=mode,
        add_deploy_dir=False,
        create_remote_dir=create_remote_dir,
    )


@operation()
def move(src: str, dest: str, overwrite=False):
    """
    Move remote file/directory/link into remote directory

    + src: remote file/directory to move
    + dest: remote directory to move `src` into
    + overwrite: whether to overwrite dest, if present
    """

    if host.get_fact(File, src) is None:
        raise OperationError("src {0} does not exist".format(src))

    if not host.get_fact(Directory, dest):
        raise OperationError("dest {0} is not an existing directory".format(dest))

    full_dest_path = os.path.join(dest, os.path.basename(src))
    if host.get_fact(File, full_dest_path) is not None:
        if overwrite:
            yield StringCommand("rm", "-rf", QuoteString(full_dest_path))
        else:
            raise OperationError(
                "dest {0} already exists and `overwrite` is unset".format(full_dest_path)
            )

    yield StringCommand("mv", QuoteString(src), QuoteString(dest))


@operation()
def copy(src: str, dest: str, overwrite=False):
    """
    Copy remote file/directory/link into remote directory

    + src: remote file/directory to copy
    + dest: remote directory to copy `src` into
    + overwrite: whether to overwrite dest, if present
    """
    src_is_dir = host.get_fact(Directory, src)
    if not host.get_fact(File, src) and not src_is_dir:
        raise OperationError(f"src {src} does not exist")

    if not host.get_fact(Directory, dest):
        raise OperationError(f"dest {dest} is not an existing directory")

    dest_file_path = os.path.join(dest, os.path.basename(src))
    dest_file_exists = host.get_fact(File, dest_file_path)
    if dest_file_exists and not overwrite:
        if _remote_file_equal(src, dest_file_path):
            host.noop(f"{dest_file_path} already exists")
            return
        else:
            raise OperationError(f"{dest_file_path} already exists and is different than src")

    cp_cmd = ["cp -r"]

    if overwrite:
        cp_cmd.append("-f")

    yield StringCommand(*cp_cmd, QuoteString(src), QuoteString(dest))


def _validate_path(path):
    try:
        return os.fspath(path)
    except TypeError:
        raise OperationTypeError("`path` must be a string or `os.PathLike` object")


def _raise_or_remove_invalid_path(fs_type, path, force, force_backup, force_backup_dir):
    if force:
        if force_backup:
            backup_path = "{0}.{1}".format(path, get_timestamp())
            if force_backup_dir:
                backup_path = os.path.basename(backup_path)
                backup_path = "{0}/{1}".format(force_backup_dir, backup_path)
            yield StringCommand("mv", QuoteString(path), QuoteString(backup_path))
        else:
            yield StringCommand("rm", "-rf", QuoteString(path))
    else:
        raise OperationError("{0} exists and is not a {1}".format(path, fs_type))


@operation()
def link(
    path: str,
    target: str | None = None,
    present=True,
    user: str | None = None,
    group: str | None = None,
    symbolic=True,
    create_remote_dir=True,
    force=False,
    force_backup=True,
    force_backup_dir: str | None = None,
):
    """
    Add/remove/update links.

    + path: the name of the link
    + target: the file/directory the link points to
    + present: whether the link should exist
    + user: user to own the link
    + group: group to own the link
    + symbolic: whether to make a symbolic link (vs hard link)
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: if the target exists and is not a file, move or remove it and continue
    + force_backup: set to ``False`` to remove any existing non-file when ``force=True``
    + force_backup_dir: directory to move any backup to when ``force=True``

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Source changes:
        If the link exists and points to a different target, pyinfra will remove it and
        recreate a new one pointing to then new target.

    **Examples:**

    .. code:: python

        files.link(
            name="Create link /etc/issue2 that points to /etc/issue",
            path="/etc/issue2",
            target="/etc/issue",
        )
    """

    path = _validate_path(path)

    if present and not target:
        raise OperationError("If present is True target must be provided")

    info = host.get_fact(Link, path=path)

    if info is False:  # not a link
        yield from _raise_or_remove_invalid_path(
            "link",
            path,
            force,
            force_backup,
            force_backup_dir,
        )
        info = None

    add_args = ["ln"]
    if symbolic:
        add_args.append("-s")

    remove_cmd = StringCommand("rm", "-f", QuoteString(path))

    if not present:
        if info:
            yield remove_cmd
        else:
            host.noop("link {link} does not exist")
        return

    assert target is not None  # appease typing QuoteString below
    add_cmd = StringCommand(" ".join(add_args), QuoteString(target), QuoteString(path))

    if info is None:  # create
        if create_remote_dir:
            yield from _create_remote_dir(path, user, group)

        yield add_cmd

        if user or group:
            yield file_utils.chown(path, user, group, dereference=False)

    else:  # edit
        changed = False

        # If the target is wrong, remove & recreate the link
        if not info or info["link_target"] != target:
            changed = True
            yield remove_cmd
            yield add_cmd

        # Check user/group
        if (user and info["user"] != user) or (group and info["group"] != group):
            yield file_utils.chown(path, user, group, dereference=False)
            changed = True

        if not changed:
            host.noop("link {0} already exists".format(path))


@operation()
def file(
    path: str,
    present=True,
    user: str | None = None,
    group: str | None = None,
    mode: int | str | None = None,
    touch=False,
    create_remote_dir=True,
    force=False,
    force_backup=True,
    force_backup_dir: str | None = None,
):
    """
    Add/remove/update files.

    + path: name/path of the remote file
    + present: whether the file should exist
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files as an integer, eg: 755
    + touch: whether to touch the file
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: if the target exists and is not a file, move or remove it and continue
    + force_backup: set to ``False`` to remove any existing non-file when ``force=True``
    + force_backup_dir: directory to move any backup to when ``force=True``

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    **Example:**

    .. code:: python

        # Note: The directory /tmp/secret will get created with the default umask.
        files.file(
            name="Create /tmp/secret/file",
            path="/tmp/secret/file",
            mode="600",
            user="root",
            group="root",
            touch=True,
            create_remote_dir=True,
        )
    """

    path = _validate_path(path)

    mode = ensure_mode_int(mode)
    info = host.get_fact(File, path=path)

    if info is False:  # not a file
        yield from _raise_or_remove_invalid_path(
            "file",
            path,
            force,
            force_backup,
            force_backup_dir,
        )
        info = None

    if not present:
        if info:
            yield StringCommand("rm", "-f", QuoteString(path))
        else:
            host.noop("file {0} does not exist")
        return

    if info is None:  # create
        if create_remote_dir:
            yield from _create_remote_dir(path, user, group)

        yield StringCommand("touch", QuoteString(path))

        if mode:
            yield file_utils.chmod(path, mode)
        if user or group:
            yield file_utils.chown(path, user, group)

    else:  # update
        changed = False

        if touch:
            changed = True
            yield StringCommand("touch", QuoteString(path))

        # Check mode
        if mode and (not info or info["mode"] != mode):
            yield file_utils.chmod(path, mode)
            changed = True

        # Check user/group
        if (user and info["user"] != user) or (group and info["group"] != group):
            yield file_utils.chown(path, user, group)
            changed = True

        if not changed:
            host.noop("file {0} already exists".format(path))


@operation()
def directory(
    path: str,
    present=True,
    user: str | None = None,
    group: str | None = None,
    mode: int | str | None = None,
    recursive=False,
    force=False,
    force_backup=True,
    force_backup_dir: str | None = None,
    _no_check_owner_mode=False,
    _no_fail_on_link=False,
):
    """
    Add/remove/update directories.

    + path: path of the remote folder
    + present: whether the folder should exist
    + user: user to own the folder
    + group: group to own the folder
    + mode: permissions of the folder
    + recursive: recursively apply user/group/mode
    + force: if the target exists and is not a file, move or remove it and continue
    + force_backup: set to ``False`` to remove any existing non-file when ``force=True``
    + force_backup_dir: directory to move any backup to when ``force=True``

    ``recursive``:
        Mode is only applied recursively if the base directory mode does not match
        the specified value.  User and group are both applied recursively if the
        base directory does not match either one; otherwise they are unchanged for
        the whole tree.

    **Examples:**

    .. code:: python

        files.directory(
            name="Ensure the /tmp/dir_that_we_want_removed is removed",
            path="/tmp/dir_that_we_want_removed",
            present=False,
        )

        files.directory(
            name="Ensure /web exists",
            path="/web",
            user="myweb",
            group="myweb",
        )

        # Multiple directories
        for dir in ["/netboot/tftp", "/netboot/nfs"]:
            files.directory(
                name="Ensure the directory `{}` exists".format(dir),
                path=dir,
            )
    """

    path = _validate_path(path)

    mode = ensure_mode_int(mode)
    info = host.get_fact(Directory, path=path)

    if info is False:  # not a directory
        if _no_fail_on_link and host.get_fact(Link, path=path):
            host.noop("directory {0} already exists (as a link)".format(path))
            return
        yield from _raise_or_remove_invalid_path(
            "directory",
            path,
            force,
            force_backup,
            force_backup_dir,
        )
        info = None

    if not present:
        if info:
            yield StringCommand("rm", "-rf", QuoteString(path))
        else:
            host.noop("directory {0} does not exist")
        return

    if info is None:  # create
        yield StringCommand("mkdir", "-p", QuoteString(path))
        if mode:
            yield file_utils.chmod(path, mode, recursive=recursive)
        if user or group:
            yield file_utils.chown(path, user, group, recursive=recursive)

    else:  # update
        if _no_check_owner_mode:
            return

        changed = False

        if mode and (not info or info["mode"] != mode):
            yield file_utils.chmod(path, mode, recursive=recursive)
            changed = True

        if (user and info["user"] != user) or (group and info["group"] != group):
            yield file_utils.chown(path, user, group, recursive=recursive)
            changed = True

        if not changed:
            host.noop("directory {0} already exists".format(path))


@operation()
def flags(path: str, flags: list[str] | None = None, present=True):
    """
    Set/clear file flags.

    + path: path of the remote folder
    + flags: a list of the file flags to be set or cleared
    + present: whether the flags should be set or cleared

    **Examples:**

    .. code:: python

        files.flags(
            name="Ensure ~/Library is visible in the GUI",
            path="~/Library",
            flags="hidden",
            present=False
        )

        files.directory(
            name="Ensure no one can change these files",
            path="/something/very/important",
            flags=["uchg", "schg"],
            present=True,
            _sudo=True
        )
    """
    flags = flags or []
    if not isinstance(flags, list):
        flags = [flags]

    if len(flags) == 0:
        host.noop(f"no changes requested to flags for '{path}'")
    else:
        current_set = set(host.get_fact(Flags, path=path))
        to_change = list(set(flags) - current_set) if present else list(current_set & set(flags))

        if len(to_change) > 0:
            prefix = "" if present else "no"
            new_flags = ",".join([prefix + flag for flag in sorted(to_change)])
            yield StringCommand("chflags", new_flags, QuoteString(path))
        else:
            host.noop(
                f"'{path}' already has '{','.join(flags)}' {'set' if present else 'clear'}",
            )


@operation()
def block(
    path: str,
    content: str | list[str] | None = None,
    present=True,
    line: str | None = None,
    backup=False,
    escape_regex_characters=False,
    try_prevent_shell_expansion=False,
    before=False,
    after=False,
    marker: str | None = None,
    begin: str | None = None,
    end: str | None = None,
):
    """
    Ensure content, surrounded by the appropriate markers, is present (or not) in the file.

    + path: target remote file
    + content: what should be present in the file (between markers).
    + present: whether the content should be present in the file
    + before: should the content be added before ``line`` if it doesn't exist
    + after: should the content be added after ``line`` if it doesn't exist
    + line: regex before or after which the content should be added if it doesn't exist.
    + backup: whether to backup the file (see ``files.line``). Default False.
    + escape_regex_characters: whether to escape regex characters from the matching line
    + try_prevent_shell_expansion: tries to prevent shell expanding by values like `$`
    + marker: the base string used to mark the text.  Default is ``# {mark} PYINFRA BLOCK``
    + begin: the value for ``{mark}`` in the marker before the content. Default is ``BEGIN``
    + end: the value for ``{mark}`` in the marker after the content. Default is ``END``

    Content appended if ``line`` not found in the file
        If ``content`` is not in the file but is required (``present=True``) and ``line`` is not
        found in the file, ``content`` (surrounded by markers) will be appended to the file.  The
        file is created if necessary.

    Content prepended or appended if ``line`` not specified
        If ``content`` is not in the file but is required and ``line`` was not provided the content
        will either be  prepended to the file (if both ``before`` and ``after``
        are ``True``) or appended to the file (if both are ``False``).

    If the file is created, it is created with the default umask; otherwise the umask is preserved
    as is the owner.

    Removal ignores ``content`` and ``line``

    Preventing shell expansion works by wrapping the content in '`' before passing to `awk`.
    WARNING: This will break if the content contains raw single quotes.

    **Examples:**

    .. code:: python

        # add entry to /etc/host
        files.block(
            name="add IP address for red server",
            path="/etc/hosts",
            content="10.0.0.1 mars-one",
            before=True,
            line=".*localhost",
        )

        # have two entries in /etc/host
        files.block(
            name="add IP address for red server",
            path="/etc/hosts",
            content="10.0.0.1 mars-one\\n10.0.0.2 mars-two",
            before=True,
            line=".*localhost",
        )

        # remove marked entry from /etc/hosts
        files.block(
            name="remove all 10.* addresses from /etc/hosts",
            path="/etc/hosts",
            present=False
        )

        # add out of date warning to web page
        files.block(
            name="add out of date warning to web page",
            path="/var/www/html/something.html",
            content= "<p>Warning: this page is out of date.</p>",
            line=".*<body>.*",
            after=True
            marker="<!-- {mark} PYINFRA BLOCK -->",
        )

        # put complex alias into .zshrc
        files.block(
            path="/home/user/.zshrc",
            content="eval $(thef -a)",
            try_prevent_shell_expansion=True,
            marker="## {mark} ALIASES ##"
        )
    """

    logger.warning("The `files.block` operation is currently in beta!")

    mark_1 = (marker or MARKER_DEFAULT).format(mark=begin or MARKER_BEGIN_DEFAULT)
    mark_2 = (marker or MARKER_DEFAULT).format(mark=end or MARKER_END_DEFAULT)

    current = host.get_fact(Block, path=path, marker=marker, begin=begin, end=end)
    cmd = None

    # Use _temp_dir global argument if provided, otherwise fall back to config
    tmp_dir = None
    if host.current_op_global_arguments:
        tmp_dir = host.current_op_global_arguments.get("_temp_dir")
    if not tmp_dir:
        tmp_dir = host.get_temp_dir_config()

    # standard awk doesn't have an "in-place edit" option so we write to a tempfile and
    # if edits were successful move to dest i.e. we do: <out_prep> ... do some work ... <real_out>
    q_path = QuoteString(path)
    mode_get = (
        ""
        if current is None
        else (
            'MODE="$(stat -c %a',
            q_path,
            "2>/dev/null || stat -f %Lp",
            q_path,
            '2>/dev/null)" &&',
        )
    )
    out_prep = StringCommand(
        f'OUT="$(TMPDIR={tmp_dir} mktemp -t pyinfra.XXXXXX)" && ',
        *mode_get,
        'OWNER="$(stat -c "%u:%g"',
        q_path,
        '2>/dev/null || stat -f "%u:%g"',
        q_path,
        '2>/dev/null || echo $(id -un):$(id -gn))" &&',
    )

    mode_change = "" if current is None else ' && chmod "$MODE"'
    real_out = StringCommand(
        ' && mv "$OUT"', q_path, ' && chown "$OWNER"', q_path, mode_change, q_path
    )

    if backup and (current is not None):  # can't back up something that doesn't exist
        out_prep = StringCommand(
            "cp",
            q_path,
            QuoteString(f"{path}.{get_timestamp()}"),
            "&&",
            out_prep,
        )

    current = host.get_fact(Block, path=path, marker=marker, begin=begin, end=end)
    # None means file didn't exist, empty list means marker was not found
    cmd = None
    if present:
        if not content:
            raise OperationValueError("'content' must be supplied when 'present' == True")
        if line:
            if before == after:
                raise OperationValueError(
                    "only one of 'before' or 'after' used when 'line` is specified"
                )
        elif before != after:
            raise OperationValueError(
                "'line' must be supplied or 'before' and 'after' must be equal"
            )
        if isinstance(content, str):
            # convert string to list of lines
            content = content.split("\n")

        the_block = "\n".join([mark_1, *content, mark_2])
        if try_prevent_shell_expansion:
            the_block = f"'{the_block}'"
            if any("'" in line for line in content):
                logger.warning(
                    "content contains single quotes, shell expansion prevention may fail"
                )
        else:
            the_block = f'"{the_block}"'

        if (current is None) or ((current == []) and (before == after)):
            # a) no file or b) file but no markers and we're adding at start or end.
            # here = hex(random.randint(0, 2147483647)) # not used as not testable
            here = "PYINFRAHERE"
            original = q_path if current is not None else QuoteString("/dev/null")
            cmd = StringCommand(
                out_prep,
                "(",
                "awk '{{print}}'",
                original if not before else " - ",
                original if before else " - ",
                '> "$OUT"',
                f"<<{here}\n{the_block[1:-1]}\n{here}\n",
                ")",
                real_out,
            )
        elif current == []:  # markers not found and have a pattern to match (not start or end)
            if not isinstance(line, str):
                raise OperationTypeError("'line' must be a regex or a string")
            regex = adjust_regex(line, escape_regex_characters)
            print_before = "{ print }" if before else ""
            print_after = "{ print }" if after else ""
            prog = (
                'awk \'BEGIN {x=ARGV[2]; ARGV[2]=""} '
                f"{print_after} f!=1 && /{regex}/ {{ print x; f=1}} "
                f"END {{if (f==0) print x }} {print_before}'"
            )
            cmd = StringCommand(
                out_prep,
                prog,
                q_path,
                the_block,
                '> "$OUT"',
                real_out,
            )
        else:
            if (len(current) != len(content)) or (
                not all(lines[0] == lines[1] for lines in zip(content, current, strict=True))
            ):  # marked_block found but text is different
                prog = (
                    'awk \'BEGIN {{f=1; x=ARGV[2]; ARGV[2]=""}}'
                    f"/{mark_1}/ {{print; print x; f=0}} /{mark_2}/ {{print; f=1; next}} f'"
                )
                cmd = StringCommand(
                    out_prep,
                    prog,
                    q_path,
                    (
                        '"' + "\n".join(content) + '"'
                        if not try_prevent_shell_expansion
                        else "'" + "\n".join(content) + "'"
                    ),
                    '> "$OUT"',
                    real_out,
                )
            else:
                host.noop("content already present")

        if cmd:
            yield cmd
    else:  # remove the marked_block
        if content:
            logger.warning("'content' ignored when removing a marked_block")
        if current is None:
            host.noop("no remove required: file did not exist")
        elif current == []:
            host.noop("no remove required: markers not found")
        else:
            cmd = StringCommand(f"awk '/{mark_1}/,/{mark_2}/ {{next}} 1'")
            yield StringCommand(out_prep, cmd, q_path, "> $OUT", real_out)
